/*
 * StateElimination.h
 *
 *  Created on: 9. 2. 2014
 *      Author: tomas
 */

#ifndef STATEELIMINATION_H_
#define STATEELIMINATION_H_

#include <set>

#include <automaton/FSM/FSM.h>
#include <regexp/RegExp.h>
#include <regexp/RegExpElements.h>

#include "../interface/IConversionRE.h"
#include "../include/macros.h"

#include "RegExpOptimize.h"

namespace conversions
{

/**
 * Converts FSM to RE using State Elimination algorithm.
 * Source: Melichar 2.118
 */
class StateElimination: public IConversionRE
{
public:
    /**
     * @param fsm Source automata.
     */
    StateElimination( const automaton::FSM & fsm );

    ~StateElimination( void );

    /**
     * Performs conversion.
     * @return regular expression equivalent to source automata.
     */
    regexp::RegExp convert( void );

private:
    /**
     * Representation of extended NFA transition function (Q x R_T -> Q, R_T = regexp over alphabet T)
     */
    struct TransitionExtendedNFA
    {
        const automaton::State m_from, m_to;
        regexp::RegExpElement* m_regexp;

        /**
         * @param from From state
         * @param to To state
         * @param regexp Expression
         */
        TransitionExtendedNFA( const automaton::State & from, regexp::RegExpElement* regexp ,const automaton::State & to );
        bool operator<( const TransitionExtendedNFA & rhs ) const;
    };

    /**
     * Stores extended NFA.
     */
    struct ExtendedNFA
    {
        std::set<automaton::State> m_states, m_finalStates, m_initialStates;
        std::set<alphabet::Symbol> m_alphabet;
        std::set<TransitionExtendedNFA> m_transitions;

        const automaton::State & createUniqueState( const std::string & name );
        std::set<StateElimination::TransitionExtendedNFA> transitionsToState( const automaton::State & toState ) const;
    };

    void constructExtendedNFA( void );

    void extendExtendedNFA( void );

    /**
     * Eliminates given state
     * @param q given state
     */
    void eliminate( const automaton::State & q );

    /**
     * Cleanup
     */
    void deleteTransitions( void );

    /**
     * gets regexp representing transition from state P to state Q
     * @param p from state
     * @param q to state
     * @return RegExpElement
     */
    regexp::RegExpElement* transition( const automaton::State & p, const automaton::State & q ) const;

    /**
     * Extended NFA.
     */
    ExtendedNFA m_MR;

    /**
     * Input automaton
     */
    const automaton::FSM & m_automaton;
};

} /* namespace conversions */

#endif /* STATEELIMINATION_H_ */
