/*
 * GlushkovTraversal.cpp
 *
 *  Created on: 13. 3. 2014
 *      Author: tomas
 */

#include "GlushkovTraversal.h"

using namespace alib;
using namespace regexp;
using namespace std;

namespace conversions
{

const GlushkovSymbol& GlushkovTraversal::findSymbol( regexp::RegExpSymbol const * const symbol, const std::set<GlushkovSymbol> & symbolSet )
{
    auto it = find_if( symbolSet.begin( ), symbolSet.end( ), [ symbol ]( GlushkovSymbol const& gs ) -> bool {
        return gs.getSymbolPtr( ) == symbol;
    } );

    if( it == symbolSet.end( ) )
        throw( "GlushkovTraversal - Can not find GlushkovSymbol for regexp node." );

    return * it;
}

bool GlushkovTraversal::pos( GlushkovSymbol const& symbol, regexp::RegExpElement const * const & node )
{
    return pos( node, symbol.getSymbolPtr( ) );
}

set<GlushkovSymbol> GlushkovTraversal::first( RegExp const& re )
{
    if( re.getRegExp( ) == NULL )
        return set<GlushkovSymbol>( );

    set<GlushkovSymbol> firstSet, symbolSet = getSymbols( re );

    for( auto const& s : first( re.getRegExp( ) ) )
        firstSet.insert( findSymbol( s, symbolSet ) );

    return firstSet;
}

set<GlushkovSymbol> GlushkovTraversal::last( RegExp const& re )
{
    if( re.getRegExp( ) == NULL )
        return set<GlushkovSymbol>( );

    set<GlushkovSymbol> lastSet, symbolSet = getSymbols( re );

    for( auto const& s : last( re.getRegExp( ) ) )
        lastSet.insert( findSymbol( s, symbolSet ) );

    return lastSet;
}

set<GlushkovSymbol> GlushkovTraversal::follow( RegExp const& re, GlushkovSymbol const& symbol )
{
    if( re.getRegExp( ) == NULL )
        return set<GlushkovSymbol>( );

    set<GlushkovSymbol> followSet, symbolSet = getSymbols( re );

    for( auto const& s: follow( re.getRegExp( ), symbol.getSymbolPtr( ) ) )
        followSet.insert( findSymbol( s, symbolSet ) );

    return followSet;
}

// -----------------------------------------------------------------------------

set<RegExpSymbol const *> GlushkovTraversal::first( RegExpElement const * const & node )
{
    Alternation const * const alternation = dynamic_cast<Alternation const * const>( node );
    Concatenation const * const concatenation = dynamic_cast<Concatenation const * const>( node );
    Iteration const * const iteration = dynamic_cast<Iteration const * const>( node );
    RegExpSymbol const * const symbol = dynamic_cast<RegExpSymbol const * const>( node );
    RegExpEmpty const * const empty = dynamic_cast<RegExpEmpty const * const>( node );
    RegExpEpsilon const * const eps = dynamic_cast<RegExpEpsilon const * const>( node );

    if( alternation )
        return first( alternation );
    else if( concatenation )
        return first( concatenation );
    else if( iteration )
        return first( iteration );
    else if( eps )
        return first( eps );
    else if( empty )
        return first( empty );
    else if( symbol )
        return first( symbol );

    throw AlibException( "GlushkovTraversal::first - invalid RegExpElement node" );
}

set<RegExpSymbol const *> GlushkovTraversal::first( Alternation const * const & node )
{
    set<RegExpSymbol const *> ret, tmp;

    for( auto const& element : node->getElements( ) )
    {
        tmp = first( element );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::first( Concatenation const * const & node )
{
    set<RegExpSymbol const *> ret, tmp;

    for( auto const& element : node->getElements( ) )
    {
        tmp = first( element );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! element->containsEmptyString( ) ) // If regexp of this subtree can match epsilon, then we need to add next subtree
            break;
    }

    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::first( Iteration const * const & node )
{
    return first( node->getElement( ) );
}

set<RegExpSymbol const *> GlushkovTraversal::first( RegExpSymbol const * const & node )
{
    return set<RegExpSymbol const *> { node };
}

set<RegExpSymbol const *> GlushkovTraversal::first( RegExpEpsilon const * const & node )
{
    return set<RegExpSymbol const *>( );
}

set<RegExpSymbol const *> GlushkovTraversal::first( RegExpEmpty const * const & node )
{
    return set<RegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

set<RegExpSymbol const *> GlushkovTraversal::last( RegExpElement const * const & node )
{
    Alternation const * const alternation = dynamic_cast<Alternation const * const>( node );
    Concatenation const * const concatenation = dynamic_cast<Concatenation const * const>( node );
    Iteration const * const iteration = dynamic_cast<Iteration const * const>( node );
    RegExpSymbol const * const symbol = dynamic_cast<RegExpSymbol const * const>( node );
    RegExpEmpty const * const empty = dynamic_cast<RegExpEmpty const * const>( node );
    RegExpEpsilon const * const eps = dynamic_cast<RegExpEpsilon const * const>( node );

    if( symbol )
        return last( symbol );
    else if( alternation )
        return last( alternation );
    else if( concatenation )
        return last( concatenation );
    else if( iteration )
        return last( iteration );
    else if( eps )
        return last( eps );
    else if( empty )
        return last( empty );

     throw AlibException( "GlushkovTraversal::last - invalid RegExpElement node" );
}

set<RegExpSymbol const *> GlushkovTraversal::last( Alternation const * const & node )
{
    set<RegExpSymbol const *> ret;

    for( auto const& element : node->getElements( ) )
    {
        set<RegExpSymbol const *> tmp = last( element );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::last( Concatenation const * const & node )
{
    set<RegExpSymbol const *> ret, tmp;

    for( auto it = node->getElements( ).rbegin( ); it != node->getElements( ).rend( ) ; it ++ )
    {
        tmp = last( *it );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! ( * it )->containsEmptyString( ) )
            break;
    }

    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::last( Iteration const * const & node )
{
    return last( node->getElement( ) );
}

set<RegExpSymbol const *> GlushkovTraversal::last( RegExpSymbol const * const & node )
{
    return set<RegExpSymbol const *> { node };
}

set<RegExpSymbol const *> GlushkovTraversal::last( RegExpEpsilon const * const & node )
{
    return set<RegExpSymbol const *>( );
}

set<RegExpSymbol const *> GlushkovTraversal::last( RegExpEmpty const * const & node )
{
    return set<RegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

set<RegExpSymbol const *> GlushkovTraversal::follow( RegExpElement const * const & node, RegExpSymbol const * const & symbolptr )
{
    Alternation const * const alternation = dynamic_cast<Alternation const * const>( node );
    Concatenation const * const concatenation = dynamic_cast<Concatenation const * const>( node );
    Iteration const * const iteration = dynamic_cast<Iteration const * const>( node );
    RegExpSymbol const * const symbol = dynamic_cast<RegExpSymbol const * const>( node );
    RegExpEmpty const * const empty = dynamic_cast<RegExpEmpty const * const>( node );
    RegExpEpsilon const * const eps = dynamic_cast<RegExpEpsilon const * const>( node );

    if( alternation )
        return follow( alternation, symbolptr );

    else if( concatenation )
        return follow( concatenation, symbolptr );

    else if( iteration )
        return follow( iteration, symbolptr );

    else if( symbol )
        return follow( symbol, symbolptr );

    else if( empty )
        return follow( empty, symbolptr );

    else if( eps )
        return follow( eps, symbolptr );

    throw AlibException( "GlushkovTraversal::follow() - unknown RegExpElement node" );
}

set<RegExpSymbol const *> GlushkovTraversal::follow( Alternation const * const & node, RegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return follow( element, symbolptr );

    throw AlibException( "GlushkovTraversal::follow(Alt)" );
}

set<RegExpSymbol const *> GlushkovTraversal::follow( Concatenation const * const & node, RegExpSymbol const * const & symbolptr )
{
    set<RegExpSymbol const *> ret, tmp, lastSet;

    for( auto e = node->getElements( ).begin( ); e != node->getElements( ).end( ); e ++ )
    {
        if( ! pos( *e, symbolptr ) )
            continue;

        tmp = follow( *e, symbolptr );
        ret.insert( tmp.begin( ), tmp.end( ) );

        lastSet = last( *e );
        if( isInSet( symbolptr, lastSet ) )
        {
            for( auto f = next( e ); f != node->getElements( ).end( ); f ++ )
            {
                tmp = first( *f );
                ret.insert( tmp.begin( ), tmp.end( ) );

                if( ! ( *f )->containsEmptyString( ) )
                    break;
            }
        }
    }
    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::follow( Iteration const * const & node, RegExpSymbol const * const & symbolptr )
{
    set<RegExpSymbol const *> ret = follow( node->getElement( ), symbolptr );
    set<RegExpSymbol const *> lastSet = last( node->getElement( ) );

    if( isInSet( symbolptr, lastSet ) )
    {
        set<RegExpSymbol const *> firstSet = first( node->getElement( ) );
        ret.insert( firstSet.begin( ), firstSet.end( ) );
    }

    return ret;
}

set<RegExpSymbol const *> GlushkovTraversal::follow( RegExpSymbol const * const & node, RegExpSymbol const * const & symbolptr )
{
    return set<RegExpSymbol const *>( );
}

set<RegExpSymbol const *> GlushkovTraversal::follow( RegExpEmpty const * const & node, RegExpSymbol const * const & symbolptr )
{
    return set<RegExpSymbol const *>( );
}

set<RegExpSymbol const *> GlushkovTraversal::follow( RegExpEpsilon const * const & node, RegExpSymbol const * const & symbolptr )
{
    return set<RegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

bool GlushkovTraversal::pos( RegExpElement const * const & node, RegExpSymbol const * const & symbolptr )
{
    Alternation const * const alternation = dynamic_cast<Alternation const * const>( node );
    Concatenation const * const concatenation = dynamic_cast<Concatenation const * const>( node );
    Iteration const * const iteration = dynamic_cast<Iteration const * const>( node );
    RegExpSymbol const * const symbol = dynamic_cast<RegExpSymbol const * const>( node );
    RegExpEmpty const * const empty = dynamic_cast<RegExpEmpty const * const>( node );
    RegExpEpsilon const * const eps = dynamic_cast<RegExpEpsilon const * const>( node );

    if( alternation )
        return pos( alternation, symbolptr );

    else if( concatenation )
        return pos( concatenation, symbolptr );

    else if( iteration )
        return pos( iteration, symbolptr );

    else if( symbol )
        return pos( symbol, symbolptr );

    else if( empty )
        return pos( empty, symbolptr );

    else if( eps )
        return pos( eps, symbolptr );

    throw AlibException( "GlushkovTraversal::pos() - unknown RegExpElement node" );
}

bool GlushkovTraversal::pos( Alternation const * const & node, RegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return true;
    return false;
}

bool GlushkovTraversal::pos( Concatenation const * const & node, RegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return true;
    return false;
}

bool GlushkovTraversal::pos( Iteration const * const & node, RegExpSymbol const * const & symbolptr )
{
    return pos( node->getElement( ), symbolptr );
}

bool GlushkovTraversal::pos( RegExpSymbol const * const & node, RegExpSymbol const * const & symbolptr )
{
    return symbolptr == node;
}

bool GlushkovTraversal::pos( RegExpEmpty const * const & node, RegExpSymbol const * const & symbolptr )
{
    return false;
}

bool GlushkovTraversal::pos( RegExpEpsilon const * const & node, RegExpSymbol const * const & symbolptr )
{
    return false;
}

// ----------------------------------------------------------------------------

set<GlushkovSymbol> GlushkovTraversal::getSymbols( RegExp const& re )
{
    set<GlushkovSymbol> alphabet;
    int i = 0;

    getSymbols( re.getRegExp( ), alphabet, i );

    return alphabet;
}

void GlushkovTraversal::getSymbols( RegExpElement const * const & node, set<GlushkovSymbol> & alphabet, int & i )
{
    Alternation const * const alternation = dynamic_cast<Alternation const * const>( node );
    Concatenation const * const concatenation = dynamic_cast<Concatenation const * const>( node );
    Iteration const * const iteration = dynamic_cast<Iteration const * const>( node );
    RegExpSymbol const * const symbol = dynamic_cast<RegExpSymbol const * const>( node );
    RegExpEmpty const * const empty = dynamic_cast<RegExpEmpty const * const>( node );
    RegExpEpsilon const * const eps = dynamic_cast<RegExpEpsilon const * const>( node );

    if( symbol )
    {
        alphabet.insert( GlushkovSymbol( symbol, i ++ ) );
        return;
    }

    else if( alternation )
    {
        for( const auto & element : alternation->getElements( ) )
            getSymbols( element, alphabet, i );
        return;
    }

    else if( concatenation )
    {
        for( const auto & element : concatenation->getElements( ) )
            getSymbols( element, alphabet, i );
        return;
    }

    else if( iteration )
    {
        getSymbols( iteration->getElement( ), alphabet, i );
        return;
    }

    else if( empty )
        return;

    else if( eps )
        return;

    throw AlibException( "GlushkovTraversal::getSymbols() - unknown RegExpElement node" );
}

} /* namespace conversions */
