#include "RhdpdaDeterminizer2.h"

namespace determinization {
namespace rhdpda {


void RhdpdaDeterminizer2::divideTransitions()
{
    for (const auto& transition : this->rhdpda->getTransitions()) {
        if (transition.getPush().size() == 1) {
            this->pushTransitions.insert(transition);
        } else if (transition.getPop().size() == 1) {
            this->popTransitions.insert(transition);
        } else {
            this->internalTransitions.insert(transition);
        }
    }
}


string RhdpdaDeterminizer2::buildStackSymbolName(const SComponent& s, const RComponent& r, const Symbol& input)
{
    return RhdpdaUtils::buildStackSymbolName(s, r, input, *this->rdpda);
}


set<StateSymbolPair> RhdpdaDeterminizer2::combineStatesWithSymbol(const set<State>& states, const Symbol& symbol)
{
    return RhdpdaUtils::combineStatesWithSymbol(states, symbol);
}


const SComponent RhdpdaDeterminizer2::getSComponentWithPairsIdentity(const set<StateSymbolPair>& pairs)
{
    return RhdpdaUtils::getSComponentWithPairsIdentity(pairs);
}


const State& RhdpdaDeterminizer2::getOrCreateState(const SComponent& s, const RComponent& r)
{
    return RhdpdaUtils::getOrCreateState(s, r, this->states, *this->rdpda);
}


void RhdpdaDeterminizer2::initDeterminization()
{
    this->states.clear();
    this->internalTransitions.clear();
    this->pushTransitions.clear();
    this->popTransitions.clear();

    this->allStateSymbolPairs = RhdpdaUtils::buildAllStateSymbolPairs(*this->rhdpda);
    this->allSComponents = RhdpdaUtils::generateAllPossibleSComponents(this->allStateSymbolPairs);
    this->allRComponents = RhdpdaUtils::generateAllPossibleRComponents(this->allStateSymbolPairs);

    Utils::copyInputAlphabet(*this->rhdpda, *this->rdpda);
    this->divideTransitions();
}


RhdpdaDeterminizer2::RhdpdaDeterminizer2(PDA* rhdpda)
{
    this->rhdpda = rhdpda;
}


Automaton* RhdpdaDeterminizer2::determinize()
{
    const Symbol& BOTTOM_OF_STACK_SYMBOL = RhdpdaUtils::BOTTOM_OF_STACK_SYMBOL;

    this->rdpda = new PDA();
    this->initDeterminization();
    
    set<StateSymbolPair> initialSPairs = this->combineStatesWithSymbol(this->rhdpda->getStates(), BOTTOM_OF_STACK_SYMBOL);
    set<StateSymbolPair> initialRPairs = this->combineStatesWithSymbol(this->rhdpda->getInitialStates(), BOTTOM_OF_STACK_SYMBOL);
    const SComponent& initialS = this->getSComponentWithPairsIdentity(initialSPairs);
    const RComponent& initialR(initialRPairs);
    const State& initialState = this->getOrCreateState(initialS, initialR);
    this->rdpda->addInitialState(initialState);

    while (true) {
        StateData* unmarkedStateData = RhdpdaUtils::getUnmarkedState(this->states);
        if (unmarkedStateData == NULL) {
            break;
        }
        
        // Internal
        for (const auto& a : this->rhdpda->getInputAlphabet()) {
            SComponent sPrimed;
            SComponent& s = unmarkedStateData->s;
            for (const auto& pairOfPairs : s.pairsOfPairs) {
                const StateSymbolPair& x = pairOfPairs.pair1;
                const State& r = pairOfPairs.pair2.state;
                const Symbol& Y = pairOfPairs.pair2.symbol;
                for (const auto& transition : this->internalTransitions) {
                    if (transition.getFrom() == r && transition.getInput() == a) {
                        const State& q = transition.getTo();
                        sPrimed.pairsOfPairs.insert(PairOfPairs(x, StateSymbolPair(q, Y)));
                    }
                }
            }
            RComponent rPrimed;
            RComponent& r = unmarkedStateData->r;
            for (const auto& pair : r.pairs) {
                const State& p = pair.state;
                const Symbol& Y = pair.symbol;
                for (const auto& transition : this->internalTransitions) {
                    if (transition.getFrom() == p && transition.getInput() == a) {
                        const State& q = transition.getTo();
                        rPrimed.pairs.insert(StateSymbolPair(q, Y));
                    }
                }
            }
            if (sPrimed.pairsOfPairs.size() > 0 && rPrimed.pairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sPrimed, rPrimed);
                const TransitionPDA transition(fromState, a, toState);
                this->rdpda->addTransition(transition);
            }
        }
        
        // Push
        for (const auto& a : this->rhdpda->getInputAlphabet()) {
            SComponent sDoublePrimed = this->getSComponentWithPairsIdentity(this->allStateSymbolPairs);
            SComponent& s = unmarkedStateData->s;
            RComponent rDoublePrimed;
            RComponent& r = unmarkedStateData->r;
            for (const auto& pair : r.pairs) {
                const State& p = pair.state;
                for (const auto& transition : this->pushTransitions) {
                    if (transition.getFrom() == p && transition.getInput() == a) {
                        const State& q = transition.getTo();
                        const Symbol& Z = transition.getPush().front();
                        rDoublePrimed.pairs.insert(StateSymbolPair(q, Z));
                    }
                }
            }
            if (rDoublePrimed.pairs.size() > 0) {
                const State& fromState = unmarkedStateData->state;
                const State& toState = this->getOrCreateState(sDoublePrimed, rDoublePrimed);
                string stackSymbolName = this->buildStackSymbolName(s, r, a);
                list<Symbol> pop = {};
                list<Symbol> push = {Symbol(stackSymbolName)};
                const TransitionPDA transition(fromState, a, toState, pop, push);
                this->rdpda->addTransition(transition);
            }
        }

        // Pop
        for (const auto& a : this->rhdpda->getInputAlphabet()) {
            for (const auto& b : this->rhdpda->getInputAlphabet()) {
                SComponent& s = unmarkedStateData->s;

                set<StatesPair> update;
                for (const auto& pairOfPairs : s.pairsOfPairs) {
                    const State& r = pairOfPairs.pair1.state;
                    const Symbol& Z = pairOfPairs.pair1.symbol;
                    const State& rPrimed = pairOfPairs.pair2.state;
                    if (pairOfPairs.pair2.symbol == Z) {
                        for (const auto& pushTransition : this->pushTransitions) {
                            for (const auto& popTransition : this->popTransitions) {
                                const State& p = pushTransition.getFrom();
                                const State& q = popTransition.getTo();
                                if (pushTransition.getInput() == a && pushTransition.getTo() == r &&
                                        popTransition.getInput() == b && popTransition.getFrom() == rPrimed &&
                                        pushTransition.getPush().front() == Z && popTransition.getPop().front() == Z) {
                                    update.insert(StatesPair(p, q));
                                }
                            }
                        }
                    }
                }
                if (update.size() == 0) {
                    continue;
                }

                for (const auto& sPrimed : this->allSComponents) {
                    for (const auto& rPrimed : this->allRComponents) {
                        RComponent rDoublePrimed;
                        for (const auto& updateItem : update) {
                            const State& p = updateItem.state1;
                            const State& q = updateItem.state2;
                            for (const auto& rPrimedPair : rPrimed.pairs) {
                                if (rPrimedPair.state == p) {
                                    const Symbol& Y = rPrimedPair.symbol;
                                    rDoublePrimed.pairs.insert(StateSymbolPair(q, Y));
                                }
                            }
                        }
                        SComponent sDoublePrimed;
                        for (const auto& updateItem : update) {
                            const State& p = updateItem.state1;
                            const State& q = updateItem.state2;
                            for (const auto& sPrimedPair : sPrimed.pairsOfPairs) {
                                if (sPrimedPair.pair2.state == p) {
                                    const StateSymbolPair& x = sPrimedPair.pair1;
                                    const Symbol& Y = sPrimedPair.pair2.symbol;
                                    sDoublePrimed.pairsOfPairs.insert(PairOfPairs(x, StateSymbolPair(q, Y)));
                                }
                            }
                        }
                        if (sDoublePrimed.pairsOfPairs.size() > 0 && rDoublePrimed.pairs.size() > 0) {
                            const State& fromState = unmarkedStateData->state;
                            const State& toState = this->getOrCreateState(sDoublePrimed, rDoublePrimed);
                            string stackSymbolName = this->buildStackSymbolName(sPrimed, rPrimed, a);
                            list<Symbol> pop = {Symbol(stackSymbolName)};
                            list<Symbol> push = {};
                            const TransitionPDA transition(fromState, b, toState, pop, push);
                            this->rdpda->addTransition(transition);
                        }
                    }
                }
            }
        }
    }

    const set<State>& rhdpdaFinalStates = this->rhdpda->getFinalStates();
    for (const auto& state : this->states) {
        bool isFinalState = false;
        const RComponent& r = state.second.r;
        for (const auto& pair : r.pairs) {
            if (rhdpdaFinalStates.find(pair.state) != rhdpdaFinalStates.end()) {
                isFinalState = true;
            }
        }
        if (isFinalState) {
            this->rdpda->addFinalState(state.second.state);
        }
    }

    return this->rdpda;
}


}
}
