#ifndef RHDPDASTRUCTS_H_
#define RHDPDASTRUCTS_H_

#include <set>

using namespace automaton;

namespace determinization {
namespace rhdpda {


/**
 * Struct fot storing pair which consists of state and symbol.
 */
struct StateSymbolPair
{
    State state;
    Symbol symbol;

    StateSymbolPair(const State& state, const Symbol& symbol)
            : state(state),
              symbol(symbol) {}

    bool operator < (const StateSymbolPair& other) const
    {
        return (state < other.state) || ((state == other.state) && (symbol < other.symbol));
    }
    
    bool operator == (const StateSymbolPair& other) const
    {
        return (state == other.state) && (symbol == other.symbol);
    }
    
    bool operator != (const StateSymbolPair& other) const
    {
        return (state != other.state) || (symbol != other.symbol);
    }
};


/**
 * Struct for storing two pairs of state and symbol.
 */
struct PairOfPairs
{
    StateSymbolPair pair1;
    StateSymbolPair pair2;
    
    PairOfPairs(const StateSymbolPair& pair1, const StateSymbolPair& pair2)
            : pair1(pair1),
              pair2(pair2) {}
    
    bool operator < (const PairOfPairs& other) const
    {
        return (pair1 < other.pair1) || ((pair1 == other.pair1) && (pair2 < other.pair2));
    }
    
    bool operator == (const PairOfPairs& other) const
    {
        return (pair1 == other.pair1) && (pair2 == other.pair2);
    }
    
    bool operator != (const PairOfPairs& other) const
    {
        return (pair1 != other.pair1) || (pair2 != other.pair2);
    }
};


/**
 * Struct for storing S component of state from deterministic pda.
 */
struct SComponent
{
    set<PairOfPairs> pairsOfPairs;
    SComponent() {}
    SComponent(const set<PairOfPairs>& pairsOfPairs)
            : pairsOfPairs(pairsOfPairs) {}
};


/**
 * Struct for storing R component of state from deterministic pda.
 */
struct RComponent
{
    set<StateSymbolPair> pairs;
    RComponent() {}
    RComponent(const set<StateSymbolPair>& pairs)
            : pairs(pairs) {}
};


/**
 * Struct for storing data about state from deterministic pda.
 */
struct StateData
{

    /** State of deterministic pda */
    State state;

    /** S component of this state */
    SComponent s;

    /** R component of this state */
    RComponent r;

    /** Indicates if this state is already marked by determinization algorithm or not */
    bool isMarked;
    
    StateData(const State& state, const SComponent& s, const RComponent& r)
            : state(state),
              s(s),
              r(r),
              isMarked(false) {}
    
    StateData(const State& state, const SComponent& s)
            : state(state),
              s(s),
              isMarked(false) {}
};


/**
 * Struct fot storing pair of two states.
 */
struct StatesPair
{
    State state1;
    State state2;
    
    StatesPair(const State& state1, const State& state2)
            : state1(state1),
              state2(state2) {}
    
    bool operator < (const StatesPair& other) const
    {
        return (state1 < other.state1) || ((state1 == other.state1) && (state2 < other.state2));
    }
    
    bool operator == (const StatesPair& other) const
    {
        return (state1 == other.state1) && (state2 == other.state2);
    }
    
    bool operator != (const StatesPair& other) const
    {
        return (state1 != other.state1) || (state2 != other.state2);
    }
};


}
}
#endif
