/*
 * AutomatonPrinter.cpp
 *
 *  Created on: Nov 11, 2013
 *      Author: martin
 */

#include "AutomatonPrinter.h"

namespace automaton {

const string AutomatonPrinter::INDENTATION = "\t";

void AutomatonPrinter::printStates(const set<State>& states, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">\n";
	for (auto state : states) {
		printState(state, out, "state", prefix + INDENTATION);
	}
	out << prefix << "</" << tagName << ">\n";
}

void AutomatonPrinter::printAlphabet(const set<Symbol>& alphabet, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">\n";
	for (auto symbol : alphabet) {
		printSymbol(symbol, out, "symbol", prefix + INDENTATION);
	}
	out << prefix << "</" << tagName << ">\n";
}

void AutomatonPrinter::printUnknownTransitions(const set<UnknownTransition>& transitions, ostream& out, string prefix) {
	string contentPrefix = prefix + INDENTATION + INDENTATION;

	out << prefix << "<transitions>\n";
	for (auto transition : transitions) {
		out << prefix << INDENTATION << "<transition>\n";
		printState(transition.getFrom(), out, "from", contentPrefix);
		printSymbol(transition.getInput(), out, "input", contentPrefix);
		printState(transition.getTo(), out, "to", contentPrefix);

		if (transition.getPop().size() > 0) {
			printSymbolList(transition.getPop(), out, "pop", contentPrefix);
		}
		if (transition.getPush().size() > 0) {
			printSymbolList(transition.getPush(), out, "push", contentPrefix);
		}

		if (transition.getOutput().getSymbol() != "") {
			printSymbol(transition.getOutput(), out, "output", contentPrefix);
		}

		if (transition.getShift() != Shift::NOT_SET) {
			printShift(transition.getShift(), out, "shift", contentPrefix);
		}

		out << prefix << INDENTATION << "</transition>\n";
	}

	out << prefix << "</transitions>\n";
}

void AutomatonPrinter::printTransitionsFSM(const set<TransitionFSM>& transitions, ostream& out, string prefix) {
	string contentPrefix = prefix + INDENTATION + INDENTATION;

	out << prefix << "<transitions>\n";
	for (auto transition : transitions) {
		out << prefix << INDENTATION << "<transition>\n";
		printState(transition.getFrom(), out, "from", contentPrefix);
		printSymbol(transition.getInput(), out, "input", contentPrefix);
		printState(transition.getTo(), out, "to", contentPrefix);
		out << prefix << INDENTATION << "</transition>\n";
	}
	out << prefix << "</transitions>\n";
}

void AutomatonPrinter::printTransitionsPDA(const set<TransitionPDA>& transitions, ostream& out, string prefix) {
	string contentPrefix = prefix + INDENTATION + INDENTATION;

	out << prefix << "<transitions>\n";
	for (auto transition : transitions) {
		out << prefix << INDENTATION << "<transition>\n";
		printState(transition.getFrom(), out, "from", contentPrefix);
		printSymbol(transition.getInput(), out, "input", contentPrefix);
		printState(transition.getTo(), out, "to", contentPrefix);
		printSymbolList(transition.getPop(), out, "pop", contentPrefix);
		printSymbolList(transition.getPush(), out, "push", contentPrefix);
		out << prefix << INDENTATION << "</transition>\n";
	}
	out << prefix << "</transitions>\n";
}

void AutomatonPrinter::printTransitionsTM(const set<TransitionTM>& transitions, ostream& out, string prefix) {
	string contentPrefix = prefix + INDENTATION + INDENTATION;

	out << prefix << "<transitions>\n";
	for (auto transition : transitions) {
		out << prefix << INDENTATION << "<transition>\n";
		printState(transition.getFrom(), out, "from", contentPrefix);
		printSymbol(transition.getInput(), out, "input", contentPrefix);
		printState(transition.getTo(), out, "to", contentPrefix);
		printSymbol(transition.getOutput(), out, "output", contentPrefix);
		printShift(transition.getShift(), out, "shift", contentPrefix);
		out << prefix << INDENTATION << "</transition>\n";
	}
	out << prefix << "</transitions>\n";
}

void AutomatonPrinter::printState(const State& state, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">";
	out << state.getName();
	out << "</" << tagName << ">\n";
}

void AutomatonPrinter::printSymbol(const Symbol& symbol, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">";
	out << symbol.getSymbol();
	out << "</" << tagName << ">\n";
}

void AutomatonPrinter::printSymbolList(const list<Symbol>& symbols, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">\n";
	for (auto symbol : symbols) {
		printSymbol(symbol, out, "symbol", prefix + INDENTATION);
	}
	out << prefix << "</" << tagName << ">\n";
}

void AutomatonPrinter::printShift(const Shift& shift, ostream& out, string tagName, string prefix) {
	out << prefix << "<" << tagName << ">";
	out << (std::string[] ) { "left", "right", "none" } [shift];
	out << "</" << tagName << ">\n";
}

void AutomatonPrinter::toXML(const UnknownAutomaton& automaton, ostream& out) {
	out << "<automaton>\n";

	if (automaton.getStates().size() > 0) {
		printStates(automaton.getStates(), out, "states", INDENTATION);
	}

	if (automaton.getInputAlphabet().size() > 0) {
		printAlphabet(automaton.getInputAlphabet(), out, "inputAlphabet", INDENTATION);
	}
	if (automaton.getTapeAlphabet().size() > 0) {
		printAlphabet(automaton.getTapeAlphabet(), out, "tapeAlphabet", INDENTATION);
	}
	if (automaton.getStackAlphabet().size() > 0) {
		printAlphabet(automaton.getStackAlphabet(), out, "stackAlphabet", INDENTATION);
	}

	if (automaton.getTransitions().size() > 0) {
		printUnknownTransitions(automaton.getTransitions(), out, INDENTATION);
	}

	if (automaton.getBlankSymbol().getSymbol() != "") {
		printSymbol(automaton.getBlankSymbol(), out, "blankSymbol", INDENTATION);
	}
	if (automaton.getStartSymbols().size() > 0) {
		printSymbolList(automaton.getStartSymbols(), out, "startSymbols", INDENTATION);
	}

	if (automaton.getInitialStates().size() > 0) {
		printStates(automaton.getInitialStates(), out, "initialStates", INDENTATION);
	}
	if (automaton.getFinalStates().size() > 0) {
		printStates(automaton.getFinalStates(), out, "finalStates", INDENTATION);
	}

	out << "</automaton>\n";

}

void AutomatonPrinter::toXML(const FSM& automaton, ostream& out) {
	out << "<automaton>\n";
	printStates(automaton.getStates(), out, "states", INDENTATION);
	printAlphabet(automaton.getInputAlphabet(), out, "inputAlphabet", INDENTATION);
	printTransitionsFSM(automaton.getTransitions(), out, INDENTATION);
	printStates(automaton.getInitialStates(), out, "initialStates", INDENTATION);
	printStates(automaton.getFinalStates(), out, "finalStates", INDENTATION);
	out << "</automaton>\n";
}

void AutomatonPrinter::toXML(const PDA& automaton, ostream& out) {
	out << "<automaton>\n";
	printStates(automaton.getStates(), out, "states", INDENTATION);
	printAlphabet(automaton.getInputAlphabet(), out, "inputAlphabet", INDENTATION);
	printAlphabet(automaton.getStackAlphabet(), out, "stackAlphabet", INDENTATION);
	printTransitionsPDA(automaton.getTransitions(), out, INDENTATION);
	printSymbolList(automaton.getStartSymbols(), out, "startSymbols", INDENTATION);
	printStates(automaton.getInitialStates(), out, "initialStates", INDENTATION);
	printStates(automaton.getFinalStates(), out, "finalStates", INDENTATION);
	out << "</automaton>\n";
}

void AutomatonPrinter::toXML(const TM& automaton, ostream& out) {
	out << "<automaton>\n";
	printStates(automaton.getStates(), out, "states", INDENTATION);
	printAlphabet(automaton.getTapeAlphabet(), out, "tapeAlphabet", INDENTATION);
	printAlphabet(automaton.getInputAlphabet(), out, "inputAlphabet", INDENTATION);
	printTransitionsTM(automaton.getTransitions(), out, INDENTATION);
	printSymbol(automaton.getBlankSymbol(), out, "blankSymbol", INDENTATION);
	printStates(automaton.getInitialStates(), out, "initialStates", INDENTATION);
	printStates(automaton.getFinalStates(), out, "finalStates", INDENTATION);
	out << "</automaton>\n";
}

} /* namespace automaton */

