/*
 * Grammar.cpp
 *
 *  Created on: Nov 2, 2013
 *      Author: Martin Zak
 */

#include "Grammar.h"
#include "GrammarException.h"
#include "GrammarPrinter.h"

namespace grammar {

using namespace std;
using namespace alphabet;

Grammar::Grammar() : startSymbol(""){

}

Grammar::Grammar(const set<Symbol>& nonTerminalSymbols, const set<Symbol>& terminalSymbols, const Symbol& startSymbol) : startSymbol(startSymbol) {
	this->nonTerminalSymbols = nonTerminalSymbols;
	this->terminalSymbols = terminalSymbols;
}

Grammar::~Grammar() {
}

void Grammar::addTerminalSymbol(const Symbol& symbol) {
	if(nonTerminalSymbols.find(symbol) != nonTerminalSymbols.end()){
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" is nonterminal symbol.");
	}

	if(!terminalSymbols.insert(symbol).second) {
		throw GrammarException("Terminal symbol \"" + symbol.getSymbol() + "\" already exists.");
	}
}

void Grammar::removeTerminalSymbol(const Symbol& symbol) {
	for(set<Rule>::iterator it = rules.begin(); it != rules.end(); it++){
		if(it->containsSymbol(symbol)) {
			throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" is used in rule.");
		}
	}

	if(!terminalSymbols.erase(symbol)) {
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" isn't terminal symbol.");
	}
}

const set<Symbol>& Grammar::getTerminalSymbols() const {
	return terminalSymbols;
}

void Grammar::addNonTerminalSymbol(const Symbol& symbol) {
	if(terminalSymbols.find(symbol) != terminalSymbols.end()){
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" is terminal symbol.");
	}

	if(!nonTerminalSymbols.insert(symbol).second) {
		throw GrammarException("Nonterminal symbol \"" + symbol.getSymbol() + "\" already exists.");
	}
}

void Grammar::removeNonTerminalSymbol(const Symbol& symbol) {
	for(set<Rule>::iterator it = rules.begin(); it != rules.end(); it++){
		if(it->containsSymbol(symbol)) {
			throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" is used in rule.");
		}
	}

	if(!nonTerminalSymbols.erase(symbol)) {
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" isn't nonterminal symbol.");
	}
}

const set<Symbol>& Grammar::getNonTerminalSymbols() const {
	return nonTerminalSymbols;
}

void Grammar::addRule(const Rule& rule) {
	if (!isValidRule(rule)) {
		throw GrammarException("Rule is not valid for this grammar. " + rule.toString());
	}

	if (!rules.insert(rule).second) {
		throw GrammarException("Rule already exists. " + rule.toString());
	}
}

void Grammar::removeRule(const Rule& rule) {
	if(!rules.erase(rule)) {
		throw GrammarException("Rule doesn't exists. " + rule.toString());
	}
}

const set<Rule>& Grammar::getRules() const {
	return rules;
}

const Symbol& Grammar::getStartSymbol() const{
	return startSymbol;
}

void Grammar::setStartSymbol(const Symbol& symbol) {
	if(nonTerminalSymbols.find(symbol) == nonTerminalSymbols.end()) {
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" isn't nonterminal symbol.");
	}

	startSymbol = symbol;
}

const Symbol& Grammar::createUniqueNonTerminalSymbol(const string& name, bool integerSuffix){
	try{
		Symbol uniqueSymbol(name);
		addNonTerminalSymbol(uniqueSymbol);
		return *nonTerminalSymbols.find(uniqueSymbol);
	}
	catch(GrammarException& e){
	}

	int i = 0;
	string str = name;

	while(i < INT_MAX)
	{
		if(integerSuffix)
			str = name + std::to_string(i++);
		else
			str += '\'';

		Symbol uniqueSymbol(str);
		try{
			addNonTerminalSymbol(uniqueSymbol);
		}
		catch(GrammarException& e){
			continue;
		}

		return *nonTerminalSymbols.find(uniqueSymbol);
	}

	throw GrammarException("Could not create unique state with name " + name + "." );
}

bool Grammar::isNonTerminalOnRightSideOfAnyRule(const Symbol& symbol) const
{
	if(nonTerminalSymbols.find(symbol) == nonTerminalSymbols.end()) {
		throw GrammarException("Symbol \"" + symbol.getSymbol() + "\" isn't nonterminal symbol.");
	}

	for(const auto & rule : rules )
	{
		const auto & rs = rule.getRightSide();
		if(find(rs.begin(), rs.end(), symbol) != rs.end())
			return true;
	}

	return false;
}

void Grammar::toXML(ostream& out) const {
	GrammarPrinter::toXML(*this, out);
}

ostream& operator <<(ostream& out, const Grammar& grammar) {
	grammar.toXML(out);
	return out;
}

} /* namespace grammar */
