/*
 * Symbol.cpp
 *
 *  Created on: Apr 16, 2013
 *      Author: Jan Travnicek
 */

#include "Symbol.h"
#include "../label/NextLabel.h"
#include "LabeledSymbol.h"
#include <climits>
#include "../exception/AlibException.h"

namespace alphabet {

Symbol::Symbol(const SymbolBase& symbol) : symbol(symbol.clone()) {

}

Symbol::Symbol(SymbolBase&& symbol) : symbol(std::move(symbol).plunder()) {

}

Symbol::Symbol(const Symbol& other) : symbol(other.getSymbol().clone()) {

}

Symbol::Symbol(Symbol&& other) noexcept : symbol(other.symbol) {
	other.symbol = NULL;
}

Symbol& Symbol::operator=(const Symbol& other) {
	if(this == &other) return *this;

	delete symbol;
	symbol = other.getSymbol().clone();

	return *this;
}

Symbol& Symbol::operator=(Symbol&& other) noexcept {
	std::swap(this->symbol, other.symbol);
	return *this;
}

Symbol::~Symbol() {
	delete symbol;
}

const SymbolBase& Symbol::getSymbol() const {
	return *symbol;
}

SymbolBase& Symbol::getSymbol() {
	return *symbol;
}

void Symbol::setSymbol(const SymbolBase& symbol) {
	delete this->symbol;
	this->symbol = symbol.clone();
}

void Symbol::setSymbol(SymbolBase&& symbol) {
	delete this->symbol;
	this->symbol = std::move(symbol).plunder();
}

bool Symbol::operator<(const Symbol& other) const {
	return *symbol < *other.symbol;
}

bool Symbol::operator!=(const Symbol& other) const {
	return !(*this == other);
}

bool Symbol::operator==(const Symbol& other) const {
	return *symbol == *other.symbol;
}

std::ostream& operator<<(std::ostream& os, const Symbol& symbol) {
	os << symbol.getSymbol();
	return os;
}

Symbol::operator std::string() const {
	return (std::string) *symbol;
}

alphabet::Symbol Symbol::createUniqueSymbol(const alphabet::Symbol& base, const std::set<alphabet::Symbol>& terminalAlphabet, const std::set<alphabet::Symbol>& nonterminalAlphabet) {
	label::NextLabel nextLabelCreator;

	const alphabet::LabeledSymbol* baseSymbol = dynamic_cast<const alphabet::LabeledSymbol*>(&(base.getSymbol()));
	if(baseSymbol == NULL)
		throw exception::AlibException("Could not create unique symbol with nonlabeled base symbol " + (std::string) base.getSymbol() + "." );

	label::Label nextLabel = baseSymbol->getLabel();

	int i = 0;
	do {
		nextLabel = nextLabelCreator.nextLabel(nextLabel);
		alphabet::Symbol attempt = alphabet::Symbol(alphabet::LabeledSymbol(nextLabel));
		if(terminalAlphabet.count(attempt) == 0 && nonterminalAlphabet.count(attempt) == 0)
			return attempt;
	} while(++i < INT_MAX);

	throw exception::AlibException("Could not create unique symbol with base symbol " + (std::string) base.getSymbol() + "." );
}

} /* namespace alphabet */

