/*
 * CompactNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "CompactNFA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <ostream>
#include <algorithm>

namespace automaton {

AutomatonBase* CompactNFA::clone() const {
	return new CompactNFA(*this);
}

AutomatonBase* CompactNFA::plunder() && {
	return new CompactNFA(std::move(*this));
}

bool CompactNFA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::pair<State, string::String>, std::set<State> >::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (t->first.first == state || t->second.find(state) != t->second.end())
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool CompactNFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (std::map<std::pair<State, string::String>, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.getAlphabet().count(symbol) == 1)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool CompactNFA::addTransition(const State& from, const string::String& input, const State& to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	std::set<alphabet::Symbol> inputStringAlphabet;
	std::set_difference(inputAlphabet.begin(), inputAlphabet.end(), input.getAlphabet().begin(), input.getAlphabet().end(), std::inserter(inputStringAlphabet, inputStringAlphabet.end()));
	if (inputStringAlphabet.size() != 0)
		throw AutomatonException("Input string is over different alphabet than automaton");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, string::String> key = std::make_pair(from, input);
	
	return transitions[key].insert(to).second;
}

bool CompactNFA::removeTransition(const State& from, const string::String& input, const State& to) {
	std::pair<State, string::String> key = std::make_pair(from, input);
	
	return transitions[key].erase(to);
}

const std::map<std::pair<State, string::String>, std::set<State>>& CompactNFA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, string::String>, std::set<State>> CompactNFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, string::String>, std::set<State>> transitionsFromState;
	for (std::map<std::pair<State, string::String>, std::set<State>>::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.first == from) {
			transitionsFromState.insert(make_pair(transition->first, transition->second));
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, string::String>, std::set<State>> CompactNFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, string::String>, std::set<State>> transitionsToState;
	for (std::map<std::pair<State, string::String>, std::set<State>>::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.find(to) != transition->second.end()) {
			transitionsToState.insert(make_pair(transition->first, transition->second));
		}
	}

	return transitionsToState;
}

bool CompactNFA::operator==(const AutomatonBase& other) const {
	return other == *this;
}

bool CompactNFA::operator==(const CompactNFA& other) const {
	return states == other.states && inputAlphabet == other.inputAlphabet && initialStates == other.initialStates && finalStates == other.finalStates && transitions == other.transitions;
}

void CompactNFA::operator>>(std::ostream& out) const {
	out << "(CompactNFA"
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialStates = " << initialStates
		<< "finalStates = " << finalStates
		<< "transitions = " << transitions
		<< ")";
}

} /* namespace automaton */
