/*
 * DFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "DFA.h"
#include "../../std/set.hpp"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <ostream>

namespace automaton {

DFA::DFA(const State& initialState) : SingleInitialState(initialState) {

}

AutomatonBase* DFA::clone() const {
	return new DFA(*this);
}

AutomatonBase* DFA::plunder() && {
	return new DFA(std::move(*this));
}

bool DFA::removeState(const State& state) {
	if (initialState == state) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::pair<State, alphabet::Symbol>, State>::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (t->first.first == state || t->second == state)
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool DFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (std::map<std::pair<State, alphabet::Symbol>, State>::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second == symbol)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);

}

bool DFA::addTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	if (inputAlphabet.find(input) == inputAlphabet.end())
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);
	
	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == to)
			return false;
		else
			throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input + "\") -> \"" + (std::string) to.getName()
						+ "\" already exists.");
	}
	
	transitions.insert(std::make_pair(key, to));
	return true;
}

bool DFA::removeTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);
	
	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& DFA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, alphabet::Symbol>, State> DFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsFromState;
	for (std::map<std::pair<State, alphabet::Symbol>, State>::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.first == from) {
			transitionsFromState.insert(std::make_pair(transition->first, transition->second));
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, alphabet::Symbol>, State> DFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, alphabet::Symbol>, State> transitionsToState;
	for (std::map<std::pair<State, alphabet::Symbol>, State>::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second == to) {
			transitionsToState.insert(std::make_pair(transition->first, transition->second));
		}
	}

	return transitionsToState;
}

bool DFA::isTotal() const {
	return transitions.size() == inputAlphabet.size() * states.size();
}

bool DFA::operator==(const AutomatonBase& other) const {
	return other == *this;
}

bool DFA::operator==(const DFA& other) const {
	return this->states == other.states && this->inputAlphabet == other.inputAlphabet && this->initialState == other.initialState && this->finalStates == other.finalStates && this->transitions == other.transitions;
}

void DFA::operator>>(std::ostream& out) const {
	out << "(DFA"
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialState = " << initialState
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

} /* namespace automaton */
