/*
 * ExtendedNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "ExtendedNFA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <ostream>
#include <algorithm>

namespace automaton {

AutomatonBase* ExtendedNFA::clone() const {
	return new ExtendedNFA(*this);
}

AutomatonBase* ExtendedNFA::plunder() && {
	return new ExtendedNFA(std::move(*this));
}

bool ExtendedNFA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::pair<State, regexp::RegExp>, std::set<State> >::const_iterator t = transitions.begin(); t != transitions.end(); t++) {
		if (t->first.first == state || t->second.find(state) != t->second.end())
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
	}

	return states.erase(state);
}

bool ExtendedNFA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (std::map<std::pair<State, regexp::RegExp>, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.second.getAlphabet().count(symbol) == 1)
			throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" is used.");
	}

	return inputAlphabet.erase(symbol);
}

bool ExtendedNFA::addTransition(const State& from, const regexp::RegExp& input, const State& to) {
	if (states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");

	std::set<alphabet::Symbol> inputRegExpAlphabet;
	std::set_difference(inputAlphabet.begin(), inputAlphabet.end(), input.getAlphabet().begin(), input.getAlphabet().end(), std::inserter(inputRegExpAlphabet, inputRegExpAlphabet.end()));
	if (inputRegExpAlphabet.size() != 0)
		throw AutomatonException("Input string is over different alphabet than automaton");

	if (states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");

	std::pair<State, regexp::RegExp> key = std::make_pair(from, input);
	
	return transitions[key].insert(to).second;
}

bool ExtendedNFA::removeTransition(const State& from, const regexp::RegExp& input, const State& to) {
	std::pair<State, regexp::RegExp> key = std::make_pair(from, input);
	
	return transitions[key].erase(to);
}

const std::map<std::pair<State, regexp::RegExp>, std::set<State> >& ExtendedNFA::getTransitions() const {
	return transitions;
}

std::map<std::pair<State, regexp::RegExp>, std::set<State> > ExtendedNFA::getTransitionsFromState(const State& from) const {
	if( states.find(from) == states.end())
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist");

	std::map<std::pair<State, regexp::RegExp>, std::set<State>> transitionsFromState;
	for (std::map<std::pair<State, regexp::RegExp>, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->first.first == from) {
			transitionsFromState.insert(make_pair(transition->first, transition->second));
		}
	}

	return transitionsFromState;
}

std::map<std::pair<State, regexp::RegExp>, std::set<State> > ExtendedNFA::getTransitionsToState(const State& to) const {
	if( states.find(to) == states.end())
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist");

	std::map<std::pair<State, regexp::RegExp>, std::set<State>> transitionsToState;
	for (std::map<std::pair<State, regexp::RegExp>, std::set<State> >::const_iterator transition = transitions.begin(); transition != transitions.end();
			transition++) {
		if (transition->second.find(to) != transition->second.end()) {
			transitionsToState.insert(make_pair(transition->first, transition->second));
		}
	}

	return transitionsToState;
}

bool ExtendedNFA::operator==(const AutomatonBase& other) const {
	return other == *this;
}

bool ExtendedNFA::operator==(const ExtendedNFA& other) const {
	return this->states == other.states && this->inputAlphabet == other.inputAlphabet && this->initialStates == other.initialStates && this->finalStates == other.finalStates && this->transitions == other.transitions;
}

void ExtendedNFA::operator>>(std::ostream& out) const {
	out << "(ExtendedNFA"
		<< " states = " << states
		<< " inputAlphabet = " << inputAlphabet
		<< " initialStates = " << initialStates
		<< " finalStates = " << finalStates
		<< " transitions = " << transitions
		<< ")";
}

} /* namespace automaton */
