/*
 * UnknownAutomaton.h
 *
 *  Created on: Oct 12, 2013
 *      Author: Martin Zak
 */

#ifndef UNKNOWN_AUTOMATON_H_
#define UNKNOWN_AUTOMATON_H_

#include <set>
#include <list>

#include "AutomatonBase.h"
#include "common/State.h"
#include "../alphabet/Symbol.h"
#include "UnknownTransition.h"

namespace automaton {

/**
 * Class representing unknown automaton parsed from XML.
 */
class UnknownAutomaton : public std::element<UnknownAutomaton, AutomatonBase> {
protected:
	std::set<State> states;
	std::set<State> initialStates;
	std::set<State> finalStates;
	std::set<alphabet::Symbol> inputAlphabet;
	std::set<alphabet::Symbol> stackAlphabet;
	std::set<alphabet::Symbol> initialSymbols;
	std::set<alphabet::Symbol> tapeAlphabet;
	alphabet::Symbol* blankSymbol;

	std::set<UnknownTransition> transitions;
public:
	UnknownAutomaton();
	~UnknownAutomaton() noexcept;

	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * Adds new state to the automaton.
	 * @param state State to add
	 * @throws AutomatonException when state already exist in the automaton
	 */
	bool addState(const State& state);

	/**
	 * Set new states of the automaton.
	 * @param states New states of the automaton
	 */
	void setStates(const std::set<State>& states);

	/**
	 * Removes the state from the automaton.
	 * @param state State to remove
	 * @throws AutomatonException when state is not present in the automaton
	 */
	bool removeState(const State& state);

	/**
	 * @return list of states
	 */
	const std::set<State>& getStates() const;


	/**
	 * Adds the State to the initial states.
	 * @param state State to add
	 * @throws AutomatonException when state is not present
	 * in the automaton or when it's already initial state
	 */
	bool addInitialState(const State& state);

	/**
	 * Sets the initial states of the automaton.
	 * @param state State to set
	 * @throws AutomatonException when state is not present
	 * in the automaton or when it's already initial state
	 */
	void setInitialStates(const std::set<State>& states);

	/**
	 * Removes the state from the initial states.
	 * @param state State to remove
	 * @throws AutomatonException when State is not initial state
	 */
	bool removeInitialState(const State& state);

	/**
	 * @return initial states
	 */
	const std::set<State>& getInitialStates() const;


	/**
	 * Adds the State to the final states.
	 * @param state State to add
	 * @throws AutomatonException when state is not present
	 * in the automaton or when it's already final state
	 */
	bool addFinalState(const State& state);

	/**
	 * Sets the final state of the automaton.
	 * @param state State to set
	 * @throws AutomatonException when state is not present
	 * in the automaton or when it's already final state
	 */
	void setFinalStates(const std::set<State>& states);

	/**
	 * Removes the state from the final states.
	 * @param state State to remove
	 * @throws AutomatonException when State is not final state
	 */
	bool removeFinalState(const State& state);

	/**
	 * @return final states
	 */
	const std::set<State>& getFinalStates() const;


	/**
	 * Adds the input symbol to the automaton.
	 * @param symbol Symbol to add
	 * @throws AutomatonException when symbol is already present in the automaton
	 */
	bool addInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * @return the input alphabet
	 */
	const std::set<alphabet::Symbol>& getInputSymbols() const;

	/**
	 * Sets the input symbols of the automaton.
	 * @param symbols Symbol to be new input symbol of the automaton
	 */
	void setInputSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * Removes input symbol from the input alphabet.
	 * @param symbol Symbol to remove
	 * @throws AutomatonException when symbol is not present in input alphabet
	 */
	bool removeInputSymbol(const alphabet::Symbol& symbol);


	/**
	 * Adds symbol to the stack alphabet.
	 * @param symbol Symbol to add
	 * @throw AutomatonException when Symbol is already present in stack alphabet
	 */
	bool addStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * Sets symbols of the stack alphabet.
	 * @param symbol Symbol to add
	 */
	void setStackSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * Removes symbol from the stack alphabet.
	 * @param symbol Symbol to remove
	 * @throw AutomatonException when Symbol is not present in input alphabet
	 */
	bool removeStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * @return the stack alphabet
	 */
	const std::set<alphabet::Symbol>& getStackSymbols() const;


	/**
	 * Adds symbol to the set of initial stack symbols.
	 * @param symbol Symbol to add
	 * @throw AutomatonException when Symbol is already present in the set of initial symbols
	 */
	bool addInitialSymbol(const alphabet::Symbol& symbol);

	/**
	 * Set the initial symbol list.
	 * @param symbols symbols to set
	 */
	void setInitialSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * Removes symbol from the set of initial stack symbols.
	 * @param symbol Symbol to remove
	 * @throw AutomatonException when Symbol is not present in the set of initial symbols
	 */
	bool removeInitialSymbol(const alphabet::Symbol& symbol);

	/**
	 * @return list of initial symbols
	 */
	const std::set<alphabet::Symbol>& getInitialSymbols() const;


	/**
	 * Adds symbol to the tape alphabet.
	 * @param symbol Symbol to add
	 * @throw AutomatonException when Symbol is already present in tape alphabet
	 */
	bool addTapeSymbol(const alphabet::Symbol& symbol);

	/**
	 * Sets the tape alphabet.
	 * @param symbols Symbols to become new tape alphabet
	 */
	void setTapeSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * Removes symbol from the tape alphabet.
	 * @param symbol Symbol to remove
	 * @throw AutomatonException when Symbol is not present in tape alphabet
	 */
	bool removeTapeSymbol(const alphabet::Symbol& symbol);

	/**
	 * @return the tape alphabet
	 */
	const std::set<alphabet::Symbol>& getTapeSymbols() const;


	/**
	 * Sets the blank symbol.
	 * @param symbol the Symbol to set
	 */
	void setBlankSymbol(const alphabet::Symbol& symbol);

	/**
	 * Tests if the automaton has blank symbol set
	 * @return true if blank symbol is set
	 */
	bool hasBlankSymbol() const;

	/**
	 * Clears setting of blank symbol
	 */
	void clearBlankSymbol();
	
	/**
	 * @return the blank symbol
	 */
	const alphabet::Symbol& getBlankSymbol() const;


	/**
	 * Adds new transition to the automaton.
	 * @param transition UnknownTransition to add
	 * @throws AutomatonException when transition is already present in the automaton
	 */
	bool addTransition(const UnknownTransition& transition);

	/**
	 * Removes the transition from the automaton.
	 * @param transition UnknownTransition to remove
	 * @throws AutomatonException when transition is not present in the automaton
	 */
	bool removeTransition(const UnknownTransition& transition);

	/**
	 * @return transitions of the automaton
	 */
	const std::set<UnknownTransition>& getTransitions() const;
	
	/**
	 * Compares two instances of UnknownAutomata
	 * @return true if this and other instance are same
	 */
	virtual bool operator==(const UnknownAutomaton& other) const;

	virtual bool operator==(const AutomatonBase& other) const;

	virtual void operator>>(std::ostream& os) const;
};

} /* namespace automaton */

#endif /* UNKNOWN_AUTOMATON_H_ */

