/*
 * UnknownTransition.h
 *
 *  Created on: Oct 13, 2013
 *      Author: Martin Zak
 */

#ifndef UNKNOWN_TRANSITION_H_
#define UNKNOWN_TRANSITION_H_

#include <vector>

#include "common/State.h"
#include "../alphabet/Symbol.h"
#include "common/Shift.h"
#include "../std/variant.hpp"
#include "../string/Epsilon.h"

namespace automaton {

/**
 * Class representing unknown transition parsed from XML. Part of UnknownAutomaton.
 */
class UnknownTransition {
protected:
	State* from;
	State* to;
	
	std::vector<alphabet::Symbol> pop;
	std::vector<alphabet::Symbol> push;

	std::variant<string::Epsilon, alphabet::Symbol>* input;
	std::variant<string::Epsilon, alphabet::Symbol>* output;

	Shift shift;

public:
	UnknownTransition();
	
	UnknownTransition(const State& from, const State& to, const std::vector<alphabet::Symbol>& pop, const std::vector<alphabet::Symbol>& push, const std::variant<string::Epsilon, alphabet::Symbol>& input, const std::variant<string::Epsilon, alphabet::Symbol>& output, Shift shift);
	~UnknownTransition();

	/**
	 * @param state State from which the transition goes
	 */
	void setFrom(const State& state);

	/**
	 * @return true if the from is set
	 */
	bool hasFrom() const;

	/**
	 * Clears the setting of from state
	 */
	void clearFrom();

	/**
	 * @return State form which the transition goes
	 */
	const State& getFrom() const;

	/**
	 * @param state State to which the transition goes
	 */
	void setTo(const State& state);

	/**
	 * @return true if the to is set
	 */
	bool hasTo() const;

	/**
	 * Clears the setting of to state
	 */
	void clearTo();

	/**
	 * @return State to which the transition goes
	 */
	const State& getTo() const;
	
	/**
	 * @param pop vector of symbols that are popped from the stack
	 */
	void setPop(const std::vector<alphabet::Symbol>& pop);

	/**
	 * @return vector of symbols that are popped from the stack
	 */
	const std::vector<alphabet::Symbol>& getPop() const;

	/**
	 * Add symbol to the end of the vector of symbols that are popped from the stack.
	 * @param symbol Symbol to add
	 */
	void addPop(const alphabet::Symbol& symbol);
	
	/**
	 * @param push vector of symbols that are pushed to the stack
	 */
	void setPush(const std::vector<alphabet::Symbol>& push);

	/**
	 * @return vector of symbols that are pushed to the stack
	 */
	const std::vector<alphabet::Symbol>& getPush() const;

	/**
	 * Add symbol to the end of the vector of symbols that are pushed to the stack.
	 * @param symbol Symbol to add
	 */
	void addPush(const alphabet::Symbol& symbol);

	/**
	 * @return the output symbol of the transition
	 */
	const std::variant<string::Epsilon, alphabet::Symbol>& getInput() const;

	/**
	 * Sets the output Symbol of the transition.
	 * @param symbol Symbol to set
	 */
	void setInput(const std::variant<string::Epsilon, alphabet::Symbol>& symbol);

	/**
	 * @return true if the input symbol is set
	 */
	bool hasInput() const;

	/**
	 * Clears the setting of input symbol
	 */
	void clearInput();

	/**
	 * @return the output symbol of the transition
	 */
	const std::variant<string::Epsilon, alphabet::Symbol>& getOutput() const;

	/**
	 * Sets the output Symbol of the transition.
	 * @param symbol Symbol to set
	 */
	void setOutput(const std::variant<string::Epsilon, alphabet::Symbol>& symbol);

	/**
	 * @return true if the output symbol is set
	 */
	bool hasOutput() const;

	/**
	 * Clears the setting of output symbol
	 */
	void clearOutput();

	/**
	 * @return direction of movement of the reading head
	 */
	const Shift& getShift() const;

	/**
	 * Sets the direction of movement of reading head.
	 * @param shift the direction
	 */
	void setShift(const Shift& shift);

	bool operator <(const UnknownTransition& other) const;
	bool operator ==(const UnknownTransition& other) const;
	bool operator !=(const UnknownTransition& other) const;

	friend std::ostream& operator<<(std::ostream&, const UnknownTransition&);
};

} /* namespace automaton */

#endif /* UNKNOWN_TRANSITION_H_ */
