/*
 * CNF.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "CNF.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

CNF::CNF(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

CNF::CNF(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* CNF::clone() const {
	return new CNF(*this);
}

GrammarBase* CNF::plunder() && {
	return new CNF(std::move(*this));
}

bool CNF::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool CNF::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool CNF::addRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	int rSize = rightHandSide.size();

	if(rSize == 0 && leftHandSide == initialSymbol) {
		for(const auto& rule : rules) {
			for(const auto& ruleRHS : rule.second) {
				if(any_of(ruleRHS.begin(), ruleRHS.end(), [&](const alphabet::Symbol& symbol) { return initialSymbol == symbol; })) {
					throw GrammarException("Initial symbol " + (std::string) initialSymbol + "used on right hand side of already existing rule");
				}
			}
		}

		generatesEpsilon = true;
		return rules[leftHandSide].insert(rightHandSide).second;
	} else if(rSize == 1) {
		if(!nonterminalAlphabet.count(leftHandSide))
			throw GrammarException("Rule must rewrite nonterminal symbol");

		if(!terminalAlphabet.count(rightHandSide[0]))
			throw GrammarException("Rule must rewrite to terminal symbol");

		return rules[leftHandSide].insert(rightHandSide).second;
	} else if(rSize == 2) {
		if(!nonterminalAlphabet.count(leftHandSide))
			throw GrammarException("Rule must rewrite nonterminal symbol");

		for(const alphabet::Symbol& symbol : rightHandSide) {
			if(nonterminalAlphabet.find(symbol) == nonterminalAlphabet.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is not a nonterminal symbol");
			if(generatesEpsilon && symbol == initialSymbol)
				throw GrammarException("Initial symbol is already allowed to be rewritten to epsilon");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		throw GrammarException("Invalid size of right hand side of the rule");
	}
}

const std::map<alphabet::Symbol, std::set<std::vector<alphabet::Symbol>>> CNF::getRules() const {
	return rules;
}

bool CNF::removeRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	int rSize = rightHandSide.size();

	if(leftHandSide == initialSymbol && rSize == 0) {
		generatesEpsilon = false;
	}
	return rules[leftHandSide].erase(rightHandSide);
}

bool CNF::operator==(const GrammarBase& other) const {
	return other == *this;
}

bool CNF::operator==(const CNF& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

void CNF::operator>>(std::ostream& out) const {
	out << "(CNF"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

} /* namespace grammar */
