/*
 * CSG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "CSG.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

CSG::CSG(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

CSG::CSG(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* CSG::clone() const {
	return new CSG(*this);
}

GrammarBase* CSG::plunder() && {
	return new CSG(std::move(*this));
}

bool CSG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool CSG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool CSG::addRule(const std::vector<alphabet::Symbol>& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	int lSize = leftHandSide.size();
	int rSize = rightHandSide.size();

	if(lSize == 1 && leftHandSide[0] == initialSymbol && rSize == 0) {
		for(const auto& rule : rules) {
			for(const auto& ruleRHS : rule.second) {
				if(any_of(ruleRHS.begin(), ruleRHS.end(), [&](const alphabet::Symbol& symbol) { return initialSymbol == symbol; })) {
					throw GrammarException("Initial symbol " + (std::string) initialSymbol + "used on right hand side of already existing rule");
				}
			}
		}

		generatesEpsilon = true;
		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		if(lSize > rSize)
			throw GrammarException("Invalid size of right hand side of a rule");

		int lContext;
		int rContext;
		for(lContext = 0; lContext < lSize - 1 && leftHandSide[lContext] == rightHandSide[lContext]; lContext++);
		for(rContext = 0; rContext < lSize - 1 && leftHandSide[lSize - rContext] == rightHandSide[rSize - rContext]; rContext++);

		if(lContext + rContext + 1 < lSize) {
			throw GrammarException("Rule must rewrite only one symbol");
		} else
		if(lContext + rContext + 1 == lSize) {
			if(!nonterminalAlphabet.count(leftHandSide[lContext - 1]) && !nonterminalAlphabet.count(leftHandSide[lSize - rContext])) throw GrammarException("Rule must rewrite nonterminal symbol");
		} else
		if(/* lContext + rContext + 1 > lSize */ std::all_of(leftHandSide.end() - rContext, leftHandSide.begin() + lContext + 1, [&](const alphabet::Symbol symbol) {return !nonterminalAlphabet.count(symbol);}))
			throw GrammarException("Rule must rewrite nonterminal symbol");


		for(const alphabet::Symbol& symbol : leftHandSide)
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end() && nonterminalAlphabet.find(symbol) == nonterminalAlphabet.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is not neither terminal nor nonterminal symbol");

		for(const alphabet::Symbol& symbol : rightHandSide) {
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end() && nonterminalAlphabet.find(symbol) == nonterminalAlphabet.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is not neither terminal nor nonterminal symbol");
			if(generatesEpsilon && symbol == initialSymbol)
				throw GrammarException("Initial symbol is already allowed to be rewritten to epsilon");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	}
}

const std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> CSG::getRules() const {
	return rules;
}

bool CSG::removeRule(const std::vector<alphabet::Symbol>& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	int lSize = leftHandSide.size();
	int rSize = rightHandSide.size();

	if(lSize == 1 && leftHandSide[0] == initialSymbol && rSize == 0) {
		generatesEpsilon = false;
	}
	return rules[leftHandSide].erase(rightHandSide);
}

bool CSG::operator==(const GrammarBase& other) const {
	return other == *this;
}

bool CSG::operator==(const CSG& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

void CSG::operator>>(std::ostream& out) const {
	out << "(CSG"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

} /* namespace grammar */
