/*
 * LeftLG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "LeftLG.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

LeftLG::LeftLG(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {

}

LeftLG::LeftLG(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* LeftLG::clone() const {
	return new LeftLG(*this);
}

GrammarBase* LeftLG::plunder() && {
	return new LeftLG(std::move(*this));
}

bool LeftLG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >> >& rule : rules) {
		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::vector<alphabet::Symbol>>()) {
				const std::vector<alphabet::Symbol>& rhs = rhsTmp.get<std::vector<alphabet::Symbol>>();

				if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
					throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
			} else {
				const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rhs = rhsTmp.get<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>();

				if(std::find(rhs.second.begin(), rhs.second.end(), symbol) != rhs.second.end())
					throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
			}
	}

	return terminalAlphabet.erase(symbol);
}

bool LeftLG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >> >& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >& rhsTmp : rule.second)
			if(rhsTmp.is<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>()) {
				const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rhs = rhsTmp.get<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>();

				if(rhs.first == symbol)
					throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
			}
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool LeftLG::addRule(const alphabet::Symbol& leftHandSide, const std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >& rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<std::vector<alphabet::Symbol>>()) {
		const std::vector<alphabet::Symbol>& rhs = rightHandSide.get<std::vector<alphabet::Symbol>>();
		for(const auto & symbol : rhs ) {
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end())
				throw GrammarException("Symbol " + (std::string) symbol.getSymbol() + " is not a terminal symbol");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rhs = rightHandSide.get<std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>>();

		if(nonterminalAlphabet.find(rhs.first) == nonterminalAlphabet.end())
			throw GrammarException("Symbol " + (std::string) rhs.first.getSymbol() + " is not a nonterminal symbol");

		for(const auto & symbol : rhs.second ) {
			if(terminalAlphabet.find(symbol) == terminalAlphabet.end())
				throw GrammarException("Symbol " + (std::string) symbol.getSymbol() + " is not a terminal symbol");
		}

		return rules[leftHandSide].insert(rightHandSide).second;
	}
}

bool LeftLG::addRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> > rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

bool LeftLG::addRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> > rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

bool LeftLG::addRawRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(rightHandSide.size() == 0) {
		return addRule(leftHandSide, rightHandSide);
	} else if(nonterminalAlphabet.count(rightHandSide[0])) {
		return addRule(leftHandSide, std::make_pair(rightHandSide[0], std::vector<alphabet::Symbol>(rightHandSide.begin() + 1, rightHandSide.end())));
	} else {
		return addRule(leftHandSide, rightHandSide);
	}
}

const std::map<alphabet::Symbol, std::set<std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >> > LeftLG::getRules() const {
	return rules;
}

bool LeftLG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> >& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool LeftLG::removeRule(const alphabet::Symbol& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool LeftLG::removeRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>>& rightHandSide) {
	std::variant<std::vector<alphabet::Symbol>, std::pair<alphabet::Symbol, std::vector<alphabet::Symbol>> > rhs(rightHandSide);
	return removeRule(leftHandSide, rhs);
}

bool LeftLG::operator==(const GrammarBase& other) const {
	return other == *this;
}

bool LeftLG::operator==(const LeftLG& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

void LeftLG::operator>>(std::ostream& out) const {
	out << "(LeftLG"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

} /* namespace grammar */
