/*
 * RightRG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "RightRG.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

RightRG::RightRG(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {

}

RightRG::RightRG(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol), generatesEpsilon(false) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* RightRG::clone() const {
	return new RightRG(*this);
}

GrammarBase* RightRG::plunder() && {
	return new RightRG(std::move(*this));
}

bool RightRG::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second) {
			if((rhs.is<alphabet::Symbol>() && rhs.get<alphabet::Symbol>() == symbol) || (rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().first == symbol))
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
		}
	}

	return terminalAlphabet.erase(symbol);
}

bool RightRG::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>>& rule : rules) {
		if(rule.first == symbol)
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rhs : rule.second)
			if(rhs.get<std::pair<alphabet::Symbol, alphabet::Symbol>>().second == symbol)
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is initial symbol.");

	return nonterminalAlphabet.erase(symbol);
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rightHandSide) {
	if(!nonterminalAlphabet.count(leftHandSide))
		throw GrammarException("Rule must rewrite nonterminal symbol");

	if(rightHandSide.is<alphabet::Symbol>()) {
		const alphabet::Symbol& rhs = rightHandSide.get<alphabet::Symbol>();

		if(!terminalAlphabet.count(rhs))
			throw GrammarException("Rule must rewrite to terminal symbol");

		return rules[leftHandSide].insert(rightHandSide).second;
	} else {
		const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = rightHandSide.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();

		if(!terminalAlphabet.count(rhs.first) || !nonterminalAlphabet.count(rhs.second))
			throw GrammarException("Rule must rewrite to terminal symbol followed by nonterminal symbol");

		return rules[leftHandSide].insert(rightHandSide).second;
	}
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const alphabet::Symbol& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

bool RightRG::addRule(const alphabet::Symbol& leftHandSide, const std::pair<alphabet::Symbol, alphabet::Symbol>& rightHandSide) {
	std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>> rhs(rightHandSide);
	return addRule(leftHandSide, rhs);
}

const std::map<alphabet::Symbol, std::set<std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>>> RightRG::getRules() const {
	return rules;
}

bool RightRG::removeRule(const alphabet::Symbol& leftHandSide, const std::variant<alphabet::Symbol, std::pair<alphabet::Symbol, alphabet::Symbol>>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

void RightRG::setGeneratesEpsilon(bool genEps) {
	generatesEpsilon = genEps;
}

bool RightRG::getGeneratesEpsilon() const {
	return generatesEpsilon;
}

bool RightRG::operator==(const GrammarBase& other) const {
	return other == *this;
}

bool RightRG::operator==(const RightRG& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules && this->generatesEpsilon == other.generatesEpsilon;
}

void RightRG::operator>>(std::ostream& out) const {
	out << "(RightRG"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< "generatesEpsilon = " << generatesEpsilon
			<< ")";
}

} /* namespace grammar */
