/*
 * UnrestrictedGrammar.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "UnrestrictedGrammar.h"
#include "../../std/map.hpp"
#include "../GrammarException.h"
#include <algorithm>

#include "../../alphabet/Symbol.h"

namespace grammar {

UnrestrictedGrammar::UnrestrictedGrammar(const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {

}

UnrestrictedGrammar::UnrestrictedGrammar(const std::set<alphabet::Symbol>& nonterminalAlphabet, const std::set<alphabet::Symbol>& terminalAlphabet, const alphabet::Symbol& initialSymbol) : TerminalNonterminalAlphabetInitialSymbol(initialSymbol) {
	setNonterminalAlphabet(nonterminalAlphabet);
	setTerminalAlphabet(terminalAlphabet);
}

GrammarBase* UnrestrictedGrammar::clone() const {
	return new UnrestrictedGrammar(*this);
}

GrammarBase* UnrestrictedGrammar::plunder() && {
	return new UnrestrictedGrammar(std::move(*this));
}

bool UnrestrictedGrammar::removeTerminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	return terminalAlphabet.erase(symbol);
}

bool UnrestrictedGrammar::removeNonterminalSymbol(const alphabet::Symbol& symbol) {
	for(const std::pair<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>>& rule : rules) {
		if(std::find(rule.first.begin(), rule.first.end(), symbol) != rule.first.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");

		for(const std::vector<alphabet::Symbol>& rhs : rule.second)
			if(std::find(rhs.begin(), rhs.end(), symbol) != rhs.end())
				throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is used in rule.");
	}

	if(initialSymbol == symbol)
		throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is initial symbol.");


	return nonterminalAlphabet.erase(symbol);
}

bool UnrestrictedGrammar::addRule(const std::vector<alphabet::Symbol>& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	if(std::all_of(leftHandSide.begin(), leftHandSide.end(), [&](const alphabet::Symbol symbol) {return !nonterminalAlphabet.count(symbol);}))
		throw GrammarException("Rule must rewrite nonterminal symbol");


	for(const alphabet::Symbol& symbol : leftHandSide)
		if(terminalAlphabet.find(symbol) == terminalAlphabet.end() && nonterminalAlphabet.find(symbol) == nonterminalAlphabet.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is not neither terminal nor nonterminal symbol");

	for(const alphabet::Symbol& symbol : rightHandSide)
		if(terminalAlphabet.find(symbol) == terminalAlphabet.end() && nonterminalAlphabet.find(symbol) == nonterminalAlphabet.end())
			throw GrammarException("Symbol \"" + (std::string) symbol.getSymbol() + "\" is not neither terminal nor nonterminal symbol");

	return rules[leftHandSide].insert(rightHandSide).second;
}

const std::map<std::vector<alphabet::Symbol>, std::set<std::vector<alphabet::Symbol>>> UnrestrictedGrammar::getRules() const {
	return rules;
}

bool UnrestrictedGrammar::removeRule(const std::vector<alphabet::Symbol>& leftHandSide, const std::vector<alphabet::Symbol>& rightHandSide) {
	return rules[leftHandSide].erase(rightHandSide);
}

bool UnrestrictedGrammar::operator==(const GrammarBase& other) const {
	return other == *this;
}

bool UnrestrictedGrammar::operator==(const UnrestrictedGrammar& other) const {
	return this->nonterminalAlphabet == other.nonterminalAlphabet && this->terminalAlphabet == other.terminalAlphabet && this->initialSymbol == other.initialSymbol && this->rules == other.rules;
}

void UnrestrictedGrammar::operator>>(std::ostream& out) const {
	out << "(UnrestrictedGrammar"
			<< "nonterminalAlphabet = " << nonterminalAlphabet
			<< "terminalAlphabet = " << terminalAlphabet
			<< "initialSymbol = " << initialSymbol
			<< "rules = " << rules
			<< ")";
}

} /* namespace grammar */
