/*
 * Alternation.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "Alternation.h"
#include "../exception/AlibException.h"

namespace regexp {

Alternation::Alternation(RegExpElement&& left, RegExpElement&& right) {
	appendElement(std::move(left));
	appendElement(std::move(right));
}

Alternation::Alternation(const RegExpElement& left, const RegExpElement& right) {
	appendElement(left);
	appendElement(right);
}

Alternation::Alternation(const Alternation& other) {
	for (const auto& element : other.elements) {
		elements.push_back(element->clone());
	}
}

Alternation::Alternation(Alternation&& other) noexcept : elements(std::move(other.elements)) {
	other.elements.clear();
	this->attachRegExp(NULL);
}

Alternation& Alternation::operator=(const Alternation& other) {
	if (this == &other) {
		return *this;
	}

	*this = Alternation(other);

	return *this;
}

Alternation& Alternation::operator=(Alternation&& other) {
	std::swap(this->elements, other.elements);
	std::swap(this->parentRegExp, other.parentRegExp);
	
	this->attachRegExp(other.parentRegExp);
	
	return *this;
}

Alternation::~Alternation() noexcept {
	for (auto element : elements) {
		delete element;
	}
	elements.clear();
}

const std::vector<const RegExpElement*> & Alternation::getElements() const {
	return * reinterpret_cast<const std::vector<const RegExpElement*> * > (&elements);
}

const std::vector<RegExpElement*> & Alternation::getElements() {
	return elements;
}

void Alternation::appendElement(const RegExpElement& element) {
	RegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	this->elements.push_back(elem);
}

void Alternation::appendElement(RegExpElement&& element) {
	RegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	this->elements.push_back(elem);
}

RegExpElement* Alternation::clone() const {
	return new Alternation(*this);
}

RegExpElement* Alternation::plunder() && {
	return new Alternation(std::move(*this));
}

bool Alternation::operator<(const RegExpElement& other) const {
	return other > *this;
}

bool Alternation::operator==(const RegExpElement& other) const {
	return other == *this;
}

bool Alternation::operator>(const RegExpElement& other) const {
	return other < *this;
}


bool Alternation::operator<(const Alternation& other) const {
	int thisSize = this->elements.size();
	int otherSize = other.elements.size();
	if(thisSize < otherSize) return true;
	if(thisSize > otherSize) return false;

	auto thisIter = this->elements.begin();
	auto otherIter = other.elements.begin();
	for(; thisIter != this->elements.end(); thisIter++, otherIter++) {
	  if(**thisIter != **otherIter) break;
	}
	if(thisIter == this->elements.end()) return false;

	return **thisIter < **otherIter;
}

bool Alternation::operator==(const Alternation& other) const {
	if(this->elements.size() != other.elements.size()) return false;

	auto thisIter = this->elements.begin();
	auto otherIter = other.elements.begin();
	for(; thisIter != this->elements.end(); thisIter++, otherIter++) {
	  if(**thisIter != **otherIter) return false;
	}

	return true;
}

void Alternation::operator>>(std::ostream& out) const {
	out << "(Alternation";
	for(const auto& e : elements)
		out << " " << *e;

	out << ")";
}

bool Alternation::testSymbol( const alphabet::Symbol & symbol ) const {
	for(const auto& child : this->elements)
		if(child->testSymbol(symbol)) return true;
	return false;
}

bool Alternation::attachRegExp(const RegExp * regexp ) {
	if(this->parentRegExp == regexp) return true;
	
	this->parentRegExp = regexp;
	for(const auto& child : this->elements)
		if(!child->attachRegExp(regexp)) return false;
	return true;
}

void Alternation::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	for(const auto& child : this->elements)
		child->computeMinimalAlphabet(alphabet);
}

bool Alternation::containsEmptyString() const {
	for(const auto& e : elements)
		if(e->containsEmptyString())
			return true;

	return false; // alternation of zero regexps is empty (this cannot happen) so by default it doesn't contain epsilon
}

bool Alternation::isEmpty() const {
	for(const auto& e : elements)
		if(!e->isEmpty())
			return false;

	return true; // alternation of zero regexps is empty (this cannot happen) but by default it is empty
}

} /* namespace regexp */
