/*
 * Concatenation.cpp
 *
 *  Created on: Nov 27, 2013
 *      Author: Martin Zak
 */

#include "Concatenation.h"
#include "../exception/AlibException.h"

namespace regexp {

Concatenation::Concatenation(RegExpElement&& left, RegExpElement&& right) {
	appendElement(std::move(left));
	appendElement(std::move(right));
}

Concatenation::Concatenation(const RegExpElement& left, const RegExpElement& right) {
	appendElement(left);
	appendElement(right);
}

Concatenation::Concatenation(const Concatenation& other) {
	for (auto element : other.elements) {
		elements.push_back(element->clone());
	}
}

Concatenation::Concatenation(Concatenation&& other) noexcept : elements(std::move(other.elements)) {
	other.elements.clear();
	this->attachRegExp(NULL);
}

Concatenation& Concatenation::operator=(const Concatenation& other) {
	if(this == &other) {
		return *this;
	}

	*this = Concatenation(other);
	
	return *this;
}

Concatenation& Concatenation::operator=(Concatenation&& other) {
	std::swap(this->elements, other.elements);
	std::swap(this->parentRegExp, other.parentRegExp);
  
	this->attachRegExp(other.parentRegExp);
	
	return *this;
}

Concatenation::~Concatenation() noexcept {
	for (auto element : elements) {
		delete element;
	}
	elements.clear();
}

const std::vector<const RegExpElement*> & Concatenation::getElements() const {
	return * reinterpret_cast<const std::vector<const RegExpElement*> * > (&elements);
}

const std::vector<RegExpElement*> & Concatenation::getElements() {
	return elements;
}

void Concatenation::appendElement(const RegExpElement& element) {
	RegExpElement* elem = element.clone();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	this->elements.push_back(elem);
}

void Concatenation::appendElement(RegExpElement&& element) {
	RegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !elem->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	this->elements.push_back(elem);
}

RegExpElement* Concatenation::clone() const {
	return new Concatenation(*this);
}

RegExpElement* Concatenation::plunder() && {
	return new Concatenation(std::move(*this));
}

bool Concatenation::operator<(const RegExpElement& other) const {
	return other > *this;
}

bool Concatenation::operator==(const RegExpElement& other) const {
	return other == *this;
}

bool Concatenation::operator>(const RegExpElement& other) const {
	return other < *this;
}


bool Concatenation::operator<(const Concatenation& other) const {
	int thisSize = this->elements.size();
	int otherSize = other.elements.size();
	if(thisSize < otherSize) return true;
	if(thisSize > otherSize) return false;

	auto thisIter = this->elements.begin();
	auto otherIter = other.elements.begin();
	for(; thisIter != this->elements.end(); thisIter++, otherIter++) {
		if(**thisIter != **otherIter) break;
	}
	if(thisIter == this->elements.end()) return false;

	return **thisIter < **otherIter;
}

bool Concatenation::operator==(const Concatenation& other) const {
	if(this->elements.size() != other.elements.size()) return false;

	auto thisIter = this->elements.begin();
	auto otherIter = other.elements.begin();
	for(; thisIter != this->elements.end(); thisIter++, otherIter++) {
		if(**thisIter != **otherIter) return false;
	}

	return true;
}

void Concatenation::operator>>(std::ostream& out) const {
	out << "(Concatenation";
	for(const auto& e : elements)
		out << " " << *e;

	out << ")";
}

bool Concatenation::testSymbol( const alphabet::Symbol & symbol ) const {
	for(const auto& child : this->elements)
		if(child->testSymbol(symbol)) return true;
	return false;
}

bool Concatenation::attachRegExp(const RegExp * regexp ) {
	if(this->parentRegExp == regexp) return true;
	
	this->parentRegExp = regexp;
	for(const auto& child : this->elements)
		if(!child->attachRegExp(regexp)) return false;
	return true;
}

void Concatenation::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	for(const auto& child : this->elements)
		child->computeMinimalAlphabet(alphabet);
}

bool Concatenation::containsEmptyString() const {
	for(const auto& e : elements)
		if( ! e->containsEmptyString())
			return false;

	return true; // concatenation of zero regexps is epsilon (this cannot happen) so by default it does contain epsilon
}

bool Concatenation::isEmpty() const {
	for(const auto& e : elements)
		if(e->isEmpty())
			return true;

	return false; // concatenation of zero regexps is epsilon (this cannot happen) but by default it isn't empty
}

} /* namespace regexp */
