/*
 * Iteration.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "Iteration.h"
#include "../exception/AlibException.h"

namespace regexp {

Iteration::Iteration(RegExpElement&& element) : element( NULL ) {
	this->setElement(std::move(element));
}

Iteration::Iteration(const RegExpElement& element) : element( NULL ) {
	this->setElement(element);
}


Iteration::Iteration(const Iteration& other) : element(other.element->clone()) {

}

Iteration::Iteration(Iteration&& other) noexcept : element(other.element) {
	other.element = NULL;
	this->attachRegExp(NULL);
}

Iteration& Iteration::operator=(const Iteration& other) {
	if (this == &other) {
		return *this;
	}

	*this = Iteration(other);

	return *this;
}

Iteration& Iteration::operator=(Iteration&& other) {
	std::swap(this->element, other.element);
	std::swap(this->parentRegExp, other.parentRegExp);
	
	this->attachRegExp(other.parentRegExp);

	return *this;
}

regexp::Iteration::~Iteration() noexcept {
	delete element;
}

const RegExpElement & Iteration::getElement() const {
	return *element;
}

RegExpElement & Iteration::getElement() {
	return *element;
}

void Iteration::setElement(const RegExpElement& element) {
	RegExpElement* elem = element.clone();
	if(this->parentRegExp && !this->element->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet.");
	delete this->element;
	this->element = elem;
}

void Iteration::setElement(RegExpElement&& element) {
	RegExpElement* elem = std::move(element).plunder();
	if(this->parentRegExp && !this->element->attachRegExp(this->parentRegExp))
		throw exception::AlibException("Input symbols not in the alphabet."); 
	delete this->element;
	this->element = elem;
}

RegExpElement* Iteration::clone() const {
	return new Iteration(*this);
}

RegExpElement* Iteration::plunder() && {
	return new Iteration(std::move(*this));
}

bool Iteration::operator<(const RegExpElement& other) const {
	return other > *this;
}

bool Iteration::operator==(const RegExpElement& other) const {
	return other == *this;
}

bool Iteration::operator>(const RegExpElement& other) const {
	return other < *this;
}

bool Iteration::operator<(const Iteration& other) const {
	return *(this->element) < *(other.element);
}

bool Iteration::operator==(const Iteration& other) const {
	return *(this->element) == *(other.element);
}

void Iteration::operator>>(std::ostream& out) const {
	out << "(RegExpIteration " << *element << ")";
}

bool Iteration::testSymbol( const alphabet::Symbol & symbol ) const {
	return element->testSymbol( symbol );
}

bool Iteration::attachRegExp(const RegExp * regexp ) {
	if(this->parentRegExp == regexp) return true;
	this->parentRegExp = regexp;
	return this->element->attachRegExp(regexp);
}

void Iteration::computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const {
	element->computeMinimalAlphabet(alphabet);
}

bool Iteration::containsEmptyString() const {
	return true;
}

bool Iteration::isEmpty() const {
	return false;
}

} /* namespace regexp */

