/*
 * RegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef REG_EXP_H_
#define REG_EXP_H_

#include <vector>
#include <list>
#include <string>
#include <set>
#include "RegExpElement.h"
#include "../std/visitor.hpp"

namespace regexp {
  
class RegExpElement;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class RegExp {
protected:
	RegExpElement* regExp;
	
	std::set<alphabet::Symbol> alphabet;

public:
	RegExp();
	RegExp(const std::set<alphabet::Symbol>& alphabet, const RegExpElement& regExp);
	RegExp(std::set<alphabet::Symbol>&& alphabet, RegExpElement&& regExp);
	explicit RegExp(const RegExpElement& regExp);
	explicit RegExp(RegExpElement&& regExp);

	/**
	 * Copy constructor.
	 * @param other RegExp to copy
	 */
	RegExp(const RegExp& other);
	RegExp(RegExp&& other) noexcept;
	RegExp& operator =(const RegExp& other);
	RegExp& operator =(RegExp&& other) noexcept;
	~RegExp() noexcept;

	/**
	 * @return Root node of the regular expression tree
	 */
	const RegExpElement& getRegExp() const;
	
	/**
	 * @return Root node of the regular expression tree
	 */
	RegExpElement& getRegExp();

	/**
	 * Sets the root node of the regular expression tree. Doesn't perform copy of the regExp param,
	 * just stores it!
	 * @param regExp root node to set
	 */
	void setRegExp(const RegExpElement& regExp);
	
	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp(RegExpElement&& regExp);
	
	/**
	 * Gets alphabet symbols used in RegExp.
	 * @return set of alphabet symbols used in regexp.
	 */
	const std::set<alphabet::Symbol>& getAlphabet() const;
	
	/**
	 * Adds symbol to the alphabet available in the regular expression
	 * @param symbol new symbol added to the alphabet
	 */
	bool addSymbolToAlphabet(const alphabet::Symbol & symbol);

	/**
	 * Sets the alphabet of the regular expression
	 * @param symbols new alphabet
	 */
	void setAlphabet(const std::set<alphabet::Symbol>& symbols);
	
	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet(const alphabet::Symbol & symbol);
	
	/**
	 * @return true if regexp represents empty language
	 */
	bool isEmpty() const;

	/**
	 * @return true if regexp matches empty string (epsilon)
	 */
	bool containsEmptyString() const;

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	friend std::ostream& operator<<(std::ostream& out, const RegExp& regexp);

	bool operator<(const RegExp&) const;
	bool operator<=(const RegExp&) const;
	bool operator==(const RegExp&) const;
	bool operator!=(const RegExp&) const;
	bool operator>(const RegExp&) const;
	bool operator>=(const RegExp&) const;
};

} /* namespace regexp */

#endif /* REG_EXP_H_ */
