/*
 * RegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef REG_EXP_ELEMENT_H_
#define REG_EXP_ELEMENT_H_

#include "../std/visitor.hpp"
#include "../alphabet/Symbol.h"
#include "RegExp.h"
#include <set>

namespace regexp {

class RegExp;
  
class Alternation;
class Concatenation;
class Iteration;
class RegExpSymbol;
class RegExpEmpty;
class RegExpEpsilon;

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
class RegExpElement : public std::elementBase<Alternation, Concatenation, Iteration, RegExpSymbol, RegExpEmpty, RegExpEpsilon> {
protected:
	/* 
	 * Parent regexp contanining this instance of RegExpElement
	 */
	const RegExp * parentRegExp;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Attaches the regexp to this instance and all its childs
	 * @param regexp parent regexp to attach as parent
	 * @return true if symbols used in regexp element are in the regexp's alphabet
	 */
	virtual bool attachRegExp ( const RegExp * regexp ) = 0;
	
	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet( std::set<alphabet::Symbol>& alphabet ) const = 0;
	
public:
	RegExpElement();
	
	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual RegExpElement* clone() const = 0;
	
	virtual RegExpElement* plunder() && = 0;

	virtual ~RegExpElement() noexcept;

	// RegExpEmpty < RegExpEpsilon < RegExpSymbol < RegExpIteration < RegExpAlternation < RegExpConcatenation
	virtual bool operator<(const RegExpElement&) const = 0;
	virtual bool operator==(const RegExpElement&) const = 0;
	virtual bool operator>(const RegExpElement&) const = 0;

	virtual bool operator>=(const RegExpElement&) const;
	virtual bool operator<=(const RegExpElement&) const;
	virtual bool operator!=(const RegExpElement&) const;

	virtual bool operator<(const Concatenation&) const;
	virtual bool operator<(const Alternation&) const;
	virtual bool operator<(const Iteration&) const;
	virtual bool operator<(const RegExpSymbol&) const;
	virtual bool operator<(const RegExpEpsilon&) const;
	virtual bool operator<(const RegExpEmpty&) const;

	virtual bool operator==(const Concatenation&) const;
	virtual bool operator==(const Alternation&) const;
	virtual bool operator==(const Iteration&) const;
	virtual bool operator==(const RegExpSymbol&) const;
	virtual bool operator==(const RegExpEpsilon&) const;
	virtual bool operator==(const RegExpEmpty&) const;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 */
	virtual void operator>>(std::ostream& out) const = 0;
	
	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	friend std::ostream& operator<<(std::ostream& out, const RegExpElement& regexp);

	/**
	 * @return true if this subtree of regexp matches empty string (epsilon)
	 */
	virtual bool containsEmptyString() const = 0;

	/**
	 * @return true if this subtree describes empty language
	 */
	virtual bool isEmpty() const = 0;
	
	friend class RegExp;
	
	friend class Alternation;
	friend class Concatenation;
	friend class Iteration;
	friend class RegExpSymbol;
	friend class RegExpEmpty;
	friend class RegExpEpsilon;
};

} /* namespace regexp */

#endif /* REG_EXP_ELEMENT_H_ */
