#include "RegExpFromStringParser.h"
#include "../sax/ParserException.h"
#include "../exception/AlibException.h"
#include "../label/StringLabel.h"

namespace regexp {

RegExpFromStringParser::RegExpFromStringParser(std::stringstream& input) : m_RegexpLexer(input), m_SymbolParser(input) {

}

RegExp RegExpFromStringParser::parse() {
	return parse(std::set<FEATURES>({FEATURES::COMPACT}));
}

RegExp RegExpFromStringParser::parse(const std::set<FEATURES>& features) {
	if(!features.count(FEATURES::COMPACT)) throw exception::AlibException();
	RegExpElement* element = this->alternation();
	RegExp regexp(std::move(*element));
	delete element;
	return regexp;
}

bool RegExpFromStringParser::first() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::EMPTY || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::LPAR) {
		return true;
	} else {
		return token.type == RegExpFromStringLexer::TokenType::ERROR && m_SymbolParser.first();
	}
}

bool RegExpFromStringParser::next() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::EMPTY || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::STAR || token.type == RegExpFromStringLexer::TokenType::RPAR || token.type == RegExpFromStringLexer::TokenType::PLUS) {
		return true;
	} else {
		return false;
	}
}

bool RegExpFromStringParser::last() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::TEOF) {
		return true;
	} else {
		return false;
	}
}

RegExpElement* RegExpFromStringParser::alternation() {
	return this->alternationCont(this->concatenation());
}

RegExpElement* RegExpFromStringParser::alternationCont(RegExpElement* left) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		next() || m_SymbolParser.first();

		try {
			RegExpElement* right = this->concatenation();
			Alternation* res = new Alternation(std::move(*left), std::move(*right));
			delete right;
			delete left;
			left = NULL;

			return this->alternationContCont(res);
		} catch (...) {
			delete left;
			throw;
		}
	} else {
		return left;
	}
}

RegExpElement* RegExpFromStringParser::alternationContCont(Alternation* res) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		next() || m_SymbolParser.first();

		try {
			RegExpElement* next = this->concatenation();
			res->appendElement(std::move(*next));
			delete next;

			return this->alternationContCont(res);
		} catch (...) {
			delete res;
			throw;
		}
	} else {
		return res;
	}
}


RegExpElement* RegExpFromStringParser::concatenation() {
	return this->concatenationCont(this->factor());
}

RegExpElement* RegExpFromStringParser::concatenationCont(RegExpElement* left) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {

		try {
			RegExpElement* right = this->factor();
			Concatenation* res = new Concatenation(std::move(*left), std::move(*right));
			delete right;
			delete left;
			left = NULL;

			return this->concatenationContCont(res);
		} catch (...) {
			delete left;
			throw;
		}
	} else {
		return left;
	}
	throw exception::AlibException("Impossible");
}

RegExpElement* RegExpFromStringParser::concatenationContCont(Concatenation* res) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {

		try {
			RegExpElement* next = this->factor();
			res->appendElement(std::move(*next));
			delete next;

			return this->concatenationContCont(res);
		} catch (...) {
			delete res;
			throw;
		}
	} else {
		return res;
	}
	throw exception::AlibException("Imúpssible");
}

RegExpElement* RegExpFromStringParser::factor() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::LPAR) {
		next() || m_SymbolParser.first();
	
		RegExpElement* base = this->alternation();
		token = m_RegexpLexer.token();
		if(token.type != RegExpFromStringLexer::TokenType::RPAR) throw exception::AlibException();
		return this->star(base);
	} else if(token.type == RegExpFromStringLexer::TokenType::EPS) {
		return this->star(new RegExpEpsilon());
	} else if(token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		return this->star(new RegExpEmpty());
	} else if(token.type == RegExpFromStringLexer::TokenType::ERROR) {
		RegExpSymbol* res = new RegExpSymbol(m_SymbolParser.parse());
		return this->star(res);
	} else {
		throw exception::AlibException();
	}
}

RegExpElement* RegExpFromStringParser::star(RegExpElement* elem) {
	next() || m_SymbolParser.first();

	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::STAR) {
		Iteration* iter = new Iteration(std::move(*elem));
		delete elem;
		return this->star(iter);
	} else {
		return elem;
	}
}

}
