/*
 * RegExpFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpFromXMLParser.h"
#include "../sax/ParserException.h"
#include "../label/StringLabel.h"

#include "../FromXMLParsers.h"

namespace regexp {

RegExp RegExpFromXMLParser::parse(std::list<sax::Token>& input) const {
	return parse(input, std::set<FEATURES>({FEATURES::COMPACT}));
}

RegExp RegExpFromXMLParser::parse(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(!features.count(FEATURES::COMPACT)) throw exception::AlibException();
	popToken(input, sax::Token::TokenType::START_ELEMENT, "regexp");
	
	RegExp regexp;
	parseAlphabet(input, regexp);
	
	RegExpElement* element = parseElement(input);
	regexp.setRegExp(std::move(*element));
	delete element;
	
	popToken(input, sax::Token::TokenType::END_ELEMENT, "regexp");

	return regexp;
}

bool RegExpFromXMLParser::first(std::list<sax::Token>& input) const {
	if(isToken(input, sax::Token::TokenType::START_ELEMENT, "regexp")) {
		return true;
	} else {
		return false;
	}
}

void RegExpFromXMLParser::parseAlphabet(std::list<sax::Token> &input, RegExp& regexp) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alphabet");
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		regexp.addSymbolToAlphabet(alib::FromXMLParsers::symbolParser.parse(input));
	}
	popToken(input, sax::Token::TokenType::END_ELEMENT, "alphabet");
}

RegExpElement* RegExpFromXMLParser::parseElement(std::list<sax::Token>& input) const {
	if (isToken(input, sax::Token::TokenType::START_ELEMENT, "empty")) {
		return parseEmpty(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		return parseEpsilon(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "iteration")) {
		return parseIteration(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "alternation")) {
		return parseAlternation(input);
	} else if (isToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation")) {
		return parseConcatenation(input);
	} else {
		return new RegExpSymbol(alib::FromXMLParsers::symbolParser.parse(input));
	}
}

Alternation* RegExpFromXMLParser::parseAlternation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "alternation");

	RegExpElement* element1 = parseElement(input);
	RegExpElement* element2 = parseElement(input);
	
	Alternation* alternation = new Alternation(std::move(*element1), std::move(*element2));
	delete element1;
	delete element2;
	
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		RegExpElement* element = parseElement(input);
		alternation->appendElement(std::move(*element));
		delete element;
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "alternation");
	return alternation;
}

Concatenation* RegExpFromXMLParser::parseConcatenation(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "concatenation");

	RegExpElement* element1 = parseElement(input);
	RegExpElement* element2 = parseElement(input);
	
	Concatenation* concatenation = new Concatenation(std::move(*element1), std::move(*element2));
	delete element1;
	delete element2;
	
	while (isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		RegExpElement* element = parseElement(input);
		concatenation->appendElement(std::move(*element));
		delete element;
	}

	popToken(input, sax::Token::TokenType::END_ELEMENT, "concatenation");
	return concatenation;

}

Iteration* RegExpFromXMLParser::parseIteration(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "iteration");

	RegExpElement* element = parseElement(input);
	Iteration* iteration = new Iteration(std::move(*element));
	delete element;

	popToken(input, sax::Token::TokenType::END_ELEMENT, "iteration");
	return iteration;
}

RegExpEpsilon* RegExpFromXMLParser::parseEpsilon(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon");

	RegExpEpsilon* epsilon = new RegExpEpsilon();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	return epsilon;
}

RegExpEmpty* RegExpFromXMLParser::parseEmpty(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "empty");

	RegExpEmpty* empty = new RegExpEmpty();

	popToken(input, sax::Token::TokenType::END_ELEMENT, "empty");
	return empty;
}

} /* namespace regexp */

