/*
 * String.cpp
 *
 *  Created on: Apr 16, 2013
 *      Author: Jan Travnicek
 */

#include "String.h"
#include "../exception/AlibException.h"
#include <algorithm>
#include "Epsilon.h"

namespace string {

String::String() : string(NULL) {
	setString(Epsilon());
}

String::String(const std::set<alphabet::Symbol>& alphabet, const StringBase& string) : string(NULL), alphabet(alphabet) {
	setString(string);
}

String::String(std::set<alphabet::Symbol>&& alphabet, StringBase&& string) : string(NULL), alphabet(alphabet) {
	setString(std::move(string));
}

String::String(const StringBase& string) : string(NULL) {
	string.computeMinimalAlphabet(alphabet);
	setString(string);
}

String::String(StringBase&& string) : string(NULL) {
	string.computeMinimalAlphabet(alphabet);
	setString(string);
}

String::String(const String& other) : string(other.getString().clone()), alphabet(other.alphabet) {
	this->string->attachString(this);
}

String::String(String&& other) noexcept : string(other.string), alphabet(std::move(other.alphabet)) {
	this->string->attachString(this);
	other.string = NULL;
}

String& String::operator=(const String& other) {
	if(this == &other) return *this;

	*this = String(other);

	return *this;
}

String& String::operator=(String&& other) noexcept {
	std::swap(this->string, other.string);
	std::swap(this->alphabet, other.alphabet);
	return *this;
}

String::~String() {
	delete string;
}

const StringBase& String::getString() const {
	return *string;
}

StringBase& String::getString() {
	return *string;
}

void String::setString(const StringBase& string) {
	delete this->string;
	this->string = string.clone();

	if(!this->string->attachString(this))
		throw exception::AlibException("String symbols not in the alphabet"); 
}

void String::setString(StringBase&& string) {
	delete this->string;
	this->string = std::move(string).plunder();

	if(!this->string->attachString(this))
		throw exception::AlibException("String symbols not in the alphabet"); 
}

const std::set<alphabet::Symbol>& String::getAlphabet() const {
	return alphabet;
}

bool String::addSymbolToAlphabet(const alphabet::Symbol & symbol) {
	return alphabet.insert(symbol).second;
}

void String::setAlphabet(const std::set<alphabet::Symbol> & symbols) {
	std::set<alphabet::Symbol> minimalAlphabet;
	this->string->computeMinimalAlphabet(minimalAlphabet);
	std::set<alphabet::Symbol> removedSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), symbols.begin(), symbols.end(), std::inserter(removedSymbols, removedSymbols.end()));

	if(removedSymbols.size() > 0)
		throw exception::AlibException("Input symbols are used.");

	this->alphabet = symbols;
}

bool String::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(this->string->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");
	
	return alphabet.erase(symbol);
}

bool String::operator<(const String& other) const {
	if(*(this->string) < *(other.string)) {
	  return true;
	} else if(*(this->string) > *(other.string)) {
	  return false;
	} else {
	  return this->alphabet < other.alphabet;
	}
}

bool String::operator!=(const String& other) const {
	return !(*this == other);
}

bool String::operator==(const String& other) const {
	return *string == *other.string && alphabet == other.alphabet;
}

std::ostream& operator<<(std::ostream& os, const String& string) {
	os << string.getString();
	return os;
}

String::operator std::string() const {
	return (std::string) *string;
}

} /* namespace string */

