#include <list>
#include "AutomatonTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "automaton/UnknownAutomaton.h"

#include "automaton/FSM/DFA.h"
#include "automaton/AutomatonFromXMLParser.h"
#include "automaton/AutomatonToXMLComposer.h"

#include "factory/AutomatonFactory.h"

#include "label/StringLabel.h"
#include "label/IntegerLabel.h"
#include "label/Label.h"
#include "alphabet/LabeledSymbol.h"

#include "automaton/FSM/FiniteAutomatonFromStringParser.h"
#include "automaton/FSM/FiniteAutomatonToStringComposer.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_REGISTRATION( AutomatonTest );

void AutomatonTest::setUp() {
}

void AutomatonTest::tearDown() {
}

void AutomatonTest::testXMLParser() {
  automaton::UnknownAutomaton automaton;
  automaton.setStates({automaton::State(label::Label(label::IntegerLabel(1))), automaton::State(label::Label(label::IntegerLabel(2))), automaton::State(label::Label(label::IntegerLabel(3)))});
  automaton.setInputSymbols({alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("a")))), alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("b"))))});
  automaton.addTransition(automaton::UnknownTransition());
  
  CPPUNIT_ASSERT( automaton == automaton );
  {
	automaton::AutomatonToXMLComposer composer;
	std::list<sax::Token> tokens = composer.compose(automaton);
	std::string tmp;
	sax::SaxComposeInterface::printMemory(tmp, tokens);
  
	std::list<sax::Token> tokens2;
	sax::SaxParseInterface::parseMemory(tmp, tokens2);
	automaton::AutomatonFromXMLParser parser;
	automaton::UnknownAutomaton automaton2 = parser.parseUnknownAutomaton(tokens2);

	CPPUNIT_ASSERT( automaton == automaton2 );
  }
  {
	std::string tmp = automaton::AutomatonFactory::toString(automaton);
	automaton::Automaton automaton2 = automaton::AutomatonFactory::fromString(tmp);

	CPPUNIT_ASSERT( automaton == automaton2.getAutomaton() );
  }
}

void AutomatonTest::testDFAParser() {
  automaton::DFA automaton(automaton::State(label::Label(label::IntegerLabel(1))));

  automaton.addState(automaton::State(label::Label(label::IntegerLabel(1))));
  automaton.addState(automaton::State(label::Label(label::IntegerLabel(2))));
  automaton.addState(automaton::State(label::Label(label::IntegerLabel(3))));
  automaton.addInputSymbol(alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("a")))));
  automaton.addInputSymbol(alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("b")))));
  
  automaton.addTransition(automaton::State(label::Label(label::IntegerLabel(1))), alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("a")))), automaton::State(label::Label(label::IntegerLabel(2))));
  automaton.addTransition(automaton::State(label::Label(label::IntegerLabel(2))), alphabet::Symbol(alphabet::LabeledSymbol(label::Label(label::StringLabel("b")))), automaton::State(label::Label(label::IntegerLabel(1))));

  automaton.addFinalState(automaton::State(label::Label(label::IntegerLabel(3))));
  
  CPPUNIT_ASSERT( automaton == automaton );
  {
	automaton::AutomatonToXMLComposer composer;
	std::list<sax::Token> tokens = composer.compose(automaton);
	std::string tmp;
	sax::SaxComposeInterface::printMemory(tmp, tokens);

	std::list<sax::Token> tokens2;
	sax::SaxParseInterface::parseMemory(tmp, tokens2);
	automaton::AutomatonFromXMLParser parser;
	automaton::DFA automaton2 = parser.parseDFA(tokens2);

	CPPUNIT_ASSERT( automaton == automaton2 );
  }
}

void AutomatonTest::FSMStringParserTest() {
	{
		std::string input = "ENFA a b c d #E\n"
				    ">0 3|4 5 1|3|4 - 2\n"
				    "1 2 - - - -\n"
				    "2 3 - - - -\n"
				    "3 - - 4 - -\n"
				    "4 - 5 - - 5\n"
				    "<5 - - - - 3\n";
		std::stringstream inputs(input);

		automaton::FiniteAutomatonFromStringParser parser(inputs);
		automaton::Automaton automaton = parser.parseValue();

		automaton::FiniteAutomatonToStringComposer composer;
		std::string output = composer.compose(automaton);
		std::stringstream outputs(output);

		CPPUNIT_ASSERT( input == output );

		automaton::FiniteAutomatonFromStringParser parser2(outputs);
		automaton::Automaton automaton2 = parser2.parseValue();

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
	{
		std::string input = "NFA a b c d\n"
				    ">0 3|4 5 1|3|4 -\n"
				    "1 2 - - -\n"
				    "2 3 - - -\n"
				    "3 - - 4 -\n"
				    "4 - 5 - -\n"
				    "<5 - - - 3\n";
		std::stringstream inputs(input);

		automaton::FiniteAutomatonFromStringParser parser(inputs);
		automaton::Automaton automaton = parser.parseValue();

		automaton::FiniteAutomatonToStringComposer composer;
		std::string output = composer.compose(automaton);
		std::stringstream outputs(output);

		CPPUNIT_ASSERT( input == output );

		automaton::FiniteAutomatonFromStringParser parser2(outputs);
		automaton::Automaton automaton2 = parser2.parseValue();

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
	{
		std::string input = "DFA a b c d\n"
				    ">0 3 5 1 -\n"
				    "1 2 - - -\n"
				    "2 3 - - -\n"
				    "3 - - 4 -\n"
				    "4 - 5 - -\n"
				    "<5 - - - 3\n";
		std::stringstream inputs(input);

		automaton::FiniteAutomatonFromStringParser parser(inputs);
		automaton::Automaton automaton = parser.parseValue();

		automaton::FiniteAutomatonToStringComposer composer;
		std::string output = composer.compose(automaton);
		std::stringstream outputs(output);

		CPPUNIT_ASSERT( input == output );

		automaton::FiniteAutomatonFromStringParser parser2(outputs);
		automaton::Automaton automaton2 = parser2.parseValue();

		CPPUNIT_ASSERT( automaton == automaton2 );
	}
}
